#!/usr/bin/env python3

from argparse import ArgumentParser
import logging
import os
import sys

import argcomplete

from .icinga1 import map_obj_type_to_class, parse_directory
from .icinga2 import (
    export_contact, export_contactgroup, detect_command_types, export_command, export_timeperiod,
    export_host, export_hostgroup, export_hostdependency, export_service, export_servicegroup,
    export_servicedependency, handle_options, icinga1_interval_length,
    icinga1_host_parents_mapping_policies, icinga1_host_parents_mapping_policy)
from .output import open_from_input_filepath, close_all


log = logging.getLogger(__name__)

parser = ArgumentParser()

parser.add_argument(
    '-v', '--verbose',
    action="store_true",
    dest="verbose",
    help="Enable verbose mode"
)

parser.add_argument(
    '-d', '--debug',
    action="store_true",
    dest="debug",
    help="Enable debug mode"
)

parser.add_argument(
    '-l', '--log-file',
    action="store",
    type=str,
    dest="logfile",
    help="Log file path"
)

parser.add_argument(
    '-i', '--input',
    action="store",
    type=str,
    dest="input",
    help="Icinga object config directory path to parse"
)

parser.add_argument(
    '-e', '--input-encoding',
    action="store",
    type=str,
    dest="input_encoding",
    help="Encoding of Icinga1 object config files (default: utf-8)",
    default='utf-8'
)

parser.add_argument(
    '-o', '--output',
    action="store",
    type=str,
    dest="output",
    help="Icinga object config directory path to parse"
)

parser.add_argument(
    '-E', '--output-encoding',
    action="store",
    type=str,
    dest="output_encoding",
    help="Encoding of Icinga2 object config files (default: utf-8)",
    default='utf-8'
)


def icinga1_object_mapping_option(value):
    parts = value.split(':')
    if len(parts) != 2:
        raise ValueError('Invalid Icinga1 object mapping option')
    obj_type = parts[0]
    if obj_type not in map_obj_type_to_class:
        raise ValueError(f'Unknown/unsupported object type {obj_type}')
    parts = parts[1].split('=')
    if len(parts) != 2:
        raise ValueError('Invalid Icinga1 object mapping option')
    return {'obj_type': obj_type, 'icinga1_name': parts[0], 'icinga2_name': parts[1]}


parser.add_argument(
    '-m', '--map-object',
    action="append",
    type=icinga1_object_mapping_option,
    dest="map_objects",
    help=(
        "Icinga1 objects name mapping for Icinga2. Could be use to map for instance "
        "'generic-debian' Icinga1 host template to Icinga2 'generic-debian-pa2' Host template. "
        "For this example, you have to specify --map-object host:generic-debian=generic-debian-pa2"
    ),
    default=[]
)

parser.add_argument(
    '-D', '--map-object-display-name',
    action="store_true",
    dest="map_objects_display_name",
    help="Also map objects display name in Icinga2.",
    default=[]
)

parser.add_argument(
    '--map-custom-macro',
    action="append",
    type=str,
    dest="map_custom_macros",
    help=(
        "Icinga1 custom macro name mapping for Icinga2. Could be use to map for instance "
        "'_NRPEPORT' Icinga1 host custom macro to Icinga2 'nrpe_port' host variable. "
        "For this example, you have to specify --map-custom-macro NRPEPORT=nrpe_port"
    ),
    default=[]
)

parser.add_argument(
    '--map-custom-user-macro-value',
    action="append",
    type=str,
    dest="map_custom_user_macro_values",
    help=(
        "Icinga1 custom user macro value mapping for Icinga2. Could be use to map for instance "
        "'$USER1$' Icinga1 custom user macro to '/usr/lib/nagios/plugins' in Icinga2 configuration."
        " For this example, you have to specify --map-custom-user-macro-value "
        "USER1=/usr/lib/nagios/plugins"),
    default=[]
)

parser.add_argument(
    '--host-parents-mapping-policy',
    choices=icinga1_host_parents_mapping_policies,
    type=str,
    dest="host_parents_mapping_policy",
    help=(
        'Mapping policy for Icinga1 host parents: in Icinga2, host parents have to be converted in '
        'host-to-host dependency. To do that, we could simply create on host-to-parents dependency '
        'by host with parents. Another way to do that, is to create an host variable that will '
        'store the host\'s parents and apply host-to-parents by looping on its variable values. '
        'This parameter permit to control the mapping method you prefer: possible values are '
        f'{" or ".join(icinga1_host_parents_mapping_policies)} '
        f'(defaut: {icinga1_host_parents_mapping_policy}). Note: with the variable policy, only '
        'the hosts variable parents will be converted and you have to manually add the apply '
        'Dependency rule (see https://icinga.com/docs/icinga-2/latest/doc/23-migrating-from-'
        'icinga-1x/#manual-config-migration-hints-for-host-parents).'
    ),
    default=icinga1_host_parents_mapping_policy
)


def objects_mapper(value):
    parts = value.split('.')
    if len(parts) < 2:
        raise ValueError('Invalid objects mapper reference')
    module = '.'.join(parts[0:-1])
    function = parts[-1]
    try:
        module = __import__(module)
    except ModuleNotFoundError as err:
        raise ValueError(f'Invalid objects mapper: python module {module} not found') \
            from err

    try:
        function = getattr(module, function)
    except NameError as err:
        raise ValueError(
            f'Invalid objects mapper: python module {module} does not contain '
            f'the function {function}.') from err

    if not callable(function):
        raise ValueError(f'Invalid objects mapper: {module}.{function} is not callable.')

    return function


parser.add_argument(
    '--objects-mappers',
    action="append",
    type=objects_mapper,
    dest="objects_mappers",
    help=(
        "Icinga1 objects mappers: you could specify name of a python function that will be call "
        "for each objects to permit custom adjustment on it (for example, add some variables). "
        "For example, you could specify 'myicinga2.add_my_vars_to_host' to call the "
        "add_my_vars_to_host of the python module myicinga2. This method will take the object "
        "reference as unique parameter, could detect the object type using the type object "
        "attribute and add some variables to the object attribute variables (dict). Multiple "
        "objects mappers could be specify by repeating this parameter."),
    default=[]
)

parser.add_argument(
    '--var-name-lowercase',
    action="store_true",
    dest="var_name_lowercase",
    help="Put Icinga1 custom macro names in lowercase in Icinga2 configuration"
)

parser.add_argument(
    '-I', '--ignore-unsupported-property',
    action="append",
    type=str,
    dest="ignore_unsupported_properties",
    help=(
        "You could use this option to inhibate warnings about specified unsupported properties. "
        "Multiple properties could be specified by repeating this parameter."),
    default=[]
)


def icinga1_exclude_object_option(value):
    parts = value.split(':')
    if len(parts) != 2:
        raise ValueError('Invalid Icinga1 object mapping option')
    obj_type = parts[0]
    if obj_type not in map_obj_type_to_class:
        raise ValueError(f'Unknown/unsupported object type {obj_type}')
    return {'obj_type': obj_type, 'icinga1_name': parts[1]}


parser.add_argument(
    '-x', '--exclude-object',
    action="append",
    type=icinga1_exclude_object_option,
    dest="excluded_objects",
    help=(
        "This parameter could be use to exclude some Icinga1 objects that will not be retrieved in "
        "Icinga2 configuration. For example, to exclude the check_ipv6_ping command, you have to "
        "specify --exclude-object command:check_ipv6_ping. Note: only the matching object will be "
        "not exported, all references to this object will be keeped and optionally map to another "
        "object if combine with -m/--map-object parameter."
    ),
    default=[]
)

parser.add_argument(
    '--icinga1-interval-length',
    type=int,
    dest="icinga1_interval_length",
    help=f'Icinga1 interval_length parameter value (default: {icinga1_interval_length})',
    default=icinga1_interval_length
)

argcomplete.autocomplete(parser)


def export_all_simple_objects(options, obj_type, export_method,
                              objects, objects_templates, disabled_objects):
    log.info("Export %s templates", obj_type)
    for name, obj in objects_templates[obj_type].items():  # pylint: disable=unused-variable
        if obj.is_excluded():
            continue
        fd = open_from_input_filepath(obj.filepath, options)
        export_method(obj, fd)

    log.info("Export %s objects", obj_type)
    for idx_or_name, obj in (  # pylint: disable=unused-variable
        objects[obj_type].items() if isinstance(objects[obj_type], dict)
        else enumerate(objects[obj_type])
    ):
        if obj.is_excluded():
            continue
        if obj.template_name:
            # Object already export
            continue
        fd = open_from_input_filepath(obj.filepath, options)
        export_method(obj, fd)

    log.info("Export disabled %s objects", obj_type)
    for idx_or_name, obj in (
        disabled_objects[obj_type].items() if isinstance(disabled_objects[obj_type], dict)
        else enumerate(disabled_objects[obj_type])
    ):
        if obj.is_excluded():
            continue
        try:
            fd = open_from_input_filepath(obj.filepath, options)
            export_method(obj, fd, disabled=True)
        except Exception:  # pylint: disable=broad-except
            log.exception(
                'Exception occurred export disabled %s object, ignore it:\n%s',
                obj_type, obj.format())


def main(argv=None):
    """ Entry point of the main script """
    options = parser.parse_args(argv)

    # Initialize logging
    if options.debug:
        loglevel = logging.DEBUG
    elif options.verbose:
        loglevel = logging.INFO
    else:
        loglevel = logging.WARNING

    logging.basicConfig(
        filename=options.logfile,
        level=loglevel,
        format='%(asctime)s - Icinga objects upgrader - %(levelname)s - %(message)s'
    )

    if not options.input:
        parser.error(
            'You must provide Icinga 1 input object config directory path using '
            '-i/--input parameter.')

    if not options.output:
        parser.error(
            'You must provide Icinga 2 output object config directory path using '
            '-o/--output parameter.')

    if not os.path.isdir(options.input):
        log.fatal('Icinga objects configuration directory not found (%s)', options.input)
        sys.exit(1)

    # Handle icinga2 options
    handle_options(options)

    objects, objects_templates, disabled_objects = parse_directory(
        options.input, objects_mappers=options.objects_mappers, encoding=options.output_encoding)

    #
    # Export icinga2 objects & template configuration
    #

    export_all_simple_objects(options, 'contact', export_contact,
                              objects, objects_templates, disabled_objects)
    export_all_simple_objects(options, 'contactgroup', export_contactgroup,
                              objects, objects_templates, disabled_objects)

    # Export all command
    log.info("Export command templates")
    for name, obj in objects_templates['command'].items():  # pylint: disable=unused-variable
        if obj.is_excluded():
            continue
        fd = open_from_input_filepath(obj.filepath, options)
        export_command(
            obj,
            detect_command_types(
                obj,
                objects,
                objects_templates
            ),
            fd
        )

    log.info("Export command objects")
    for name, obj in objects['command'].items():
        if obj.is_excluded():
            continue
        if obj.template_name:
            # Object already export
            continue
        fd = open_from_input_filepath(obj.filepath, options)
        export_command(
            obj,
            detect_command_types(
                obj,
                objects,
                objects_templates
            ),
            fd
        )

    log.info("Export disabled command objects")
    for name, obj in disabled_objects['command'].items():
        if obj.is_excluded():
            continue
        try:
            fd = open_from_input_filepath(obj.filepath, options)
            export_command(
                obj,
                detect_command_types(
                    obj,
                    objects,
                    objects_templates
                ),
                fd, disabled=True
            )
        except Exception:  # pylint: disable=broad-except
            log.exception(
                'Exception occurred export disabled command object, ignore it:\n%s',
                obj.format())

    export_all_simple_objects(options, 'timeperiod', export_timeperiod,
                              objects, objects_templates, disabled_objects)

    export_all_simple_objects(options, 'host', export_host,
                              objects, objects_templates, disabled_objects)
    export_all_simple_objects(options, 'hostgroup', export_hostgroup,
                              objects, objects_templates, disabled_objects)
    export_all_simple_objects(options, 'hostdependency', export_hostdependency,
                              objects, objects_templates, disabled_objects)

    export_all_simple_objects(options, 'service', export_service,
                              objects, objects_templates, disabled_objects)
    export_all_simple_objects(options, 'servicegroup', export_servicegroup,
                              objects, objects_templates, disabled_objects)
    export_all_simple_objects(options, 'servicedependency', export_servicedependency,
                              objects, objects_templates, disabled_objects)

    close_all()
