import logging
import os
import re
import sys

from .objects import (
    IcingaHost, IcingaHostGroup, IcingaService, IcingaServiceGroup, IcingaHostDependency,
    IcingaServiceDependency, IcingaContact, IcingaContactGroup, IcingaTimePeriod, IcingaCommand)


log = logging.getLogger(__name__)

default_encoding = 'utf-8'
define_start_line = re.compile(r'^\s*define\s*([^\s]+)\s*\{')
define_end_line = re.compile(r'^\s*\}')
property_line = re.compile(
    r'^\s*(?P<name>[^\s]+)\s+(?P<value>[^\}#;\n\r]+)\s*(?P<end_mark>\}?)\s*([#;]\s+'
    r'(?P<comment>.*)\s+)?$')
property_line_after_linereturn = re.compile(r'^\s*(?P<value>[^\}#;\n\r]+)\s+(?P<end_mark>\}?)\s*$')
timeperiod_property_line = re.compile(
    r'^\s*(?P<name>((sunday|monday|tuesday|wednesday|thursday|friday|saturday)'
    r'(\s+[0-9]+(\s+(january|february|march|april|may|june|july|august|september|october'
    r'|november|december))?)?)|([0-9]{4}\-[0-9]{2}\-[0-9]{2}(\s-\s[0-9]{4}\-[0-9]{2}\-'
    r'[0-9]{2})?)|((day|january|february|march|april|may|june|july|august|september|'
    r'october|november|december)\s+{0-9}+(\s+-\s+-?{0-9}+)?)|([0-9]+\s+(january|february|march|'
    r'april|may|june|july|august|september|october|november|december))|((january|february|march|'
    r'april|may|june|july|august|september|october|november|december)\s+[0-9]+)|)\s+'
    r'(?P<value>[0-9]{1,2}:[0-9]{1,2}\s*-\s*[0-9]{1,2}:[0-9]{1,2}(\s*,\s*[0-9]{1,2}:[0-9]{1,2}'
    r'\s*-\s*[0-9]{1,2}:[0-9]{1,2})*)\s*(?P<end_mark>\}?)\s*([#;]\s*(?P<comment>.*)\s*)?$')
commented_line = re.compile(r'^\s*#\s*(.*)$')

map_obj_type_to_class = {
    'host': {'class': IcingaHost, 'unique': True},
    'hostgroup': {'class': IcingaHostGroup, 'unique': True},
    'hostdependency': {'class': IcingaHostDependency, 'unique': False},
    'service': {'class': IcingaService, 'unique': False},
    'servicegroup': {'class': IcingaServiceGroup, 'unique': True},
    'servicedependency': {'class': IcingaServiceDependency, 'unique': False},
    'contact': {'class': IcingaContact, 'unique': True},
    'contactgroup': {'class': IcingaContactGroup, 'unique': True},
    'timeperiod': {'class': IcingaTimePeriod, 'unique': True},
    'command': {'class': IcingaCommand, 'unique': True},
}


def parse_file(path, encoding=None):
    objects = {}
    current_object = {}
    next_property_pre_comment_lines = []
    property_before_line_return = {}

    def add_object(end_line=True):
        log.debug(
            '[%s:%i] New object %s detected with properties:\n\t- %s',
            path, line_idx,
            current_object['type'],
            "\n\t- ".join([
                f'"{name}": "{info["value"]}"'
                f'{(" # " + info["comment"]) if info["comment"] else ""}'
                for name, info in current_object['properties'].items()
            ])
        )
        if current_object['type'] not in objects:
            objects[current_object['type']] = []
        if end_line:
            current_object['end_line'] = line_idx
        objects[current_object['type']].append(current_object)

    with open(path, 'r', encoding=encoding if encoding else default_encoding) as fd:
        line_idx = 0
        for line in fd.readlines():
            line_idx += 1
            log.debug("[%s:%i] Handle new line: '%s'", path, line_idx, line)

            if property_before_line_return:
                m = property_line_after_linereturn.match(line)
                assert m, (
                    f'[{path}:{line_idx}] Line not match with the regexp of property lines '
                    'after line return')
                property_name = property_before_line_return['name']
                line_return = m.group('value').endswith('\\')
                property_value = m.group('value').strip()
                if line_return:
                    property_value = property_value[0:-1].strip()
                end_object = m.group('end_mark')

                log.debug(
                    '[%s:%i] Property "%s" (after line return) = "%s" '
                    '(line return: %s, end object mark: %s)',
                    path, line_idx,
                    property_name, property_value,
                    'yes' if line_return else 'no',
                    'yes' if end_object else 'no'
                )

                assert not (line_return and end_object), (
                    f'[{path}:{line_idx}] Can not have line return mark and end object mark '
                    'on same line'
                )

                property_before_line_return['value'] = (
                    property_value if not property_before_line_return['value']
                    else f'{property_before_line_return["value"]} {property_value}'
                )

                if line_return:
                    continue

                current_object['properties'][property_name] = dict(
                    value=property_before_line_return['value'],
                    pre_comment_lines=next_property_pre_comment_lines,
                    comment=None,
                )
                property_before_line_return = {}
                next_property_pre_comment_lines = []
                if end_object:
                    add_object()
                    current_object = {}
                continue

            if not line.strip():
                log.debug(
                    "[%s:%i] Empty line detected, ignored",
                    path, line_idx)
                continue

            m = commented_line.match(line)
            if m:
                if current_object:
                    log.debug(
                        '[%s:%i] Commented line detected in object definition, keep it as '
                        'pre-comment lines for the next property',
                        path, line_idx
                    )
                    next_property_pre_comment_lines.append(m.group(1))
                else:
                    log.debug(
                        '[%s:%i] Commented line detected, ignored',
                        path, line_idx
                    )
                continue

            m = define_start_line.match(line)
            if m:
                # New object
                if current_object:
                    log.warning(
                        '[%s:%i] Object not properly closed: %s',
                        path, line_idx, current_object
                    )
                    add_object(end_line=False)
                log.debug(
                    '[%s:%i] New object %s detected',
                    path, line_idx, m.group(1)
                )
                current_object = {
                    'type': m.group(1),
                    'filepath': path,
                    'start_line': line_idx,
                    'end_line': None,
                    'properties': {},
                }
                next_property_pre_comment_lines = []
                continue

            m = define_end_line.match(line)
            if m:
                # Object end
                if not current_object:
                    log.error(
                        'End object at line %s:%i, but no object currently opened',
                        path, line_idx
                    )
                    continue
                add_object()
                current_object = None
                continue

            if current_object and current_object['type'] == 'timeperiod':
                m = timeperiod_property_line.match(line)
                if m:
                    # Object property line
                    if not current_object:
                        log.warning(
                            '[%s:%i] Timeperiod object property line detected, '
                            'but no object currently opened', path, line_idx)
                        continue
                    property_name = m.group('name')
                    property_value = m.group('value').strip()
                    end_object = m.group('end_mark')
                    comment = m.group('comment')
                    log.debug(
                        '[%s:%i] Timeperiod property "%s" = "%s" (end object mark: %s)',
                        path, line_idx,
                        property_name, property_value,
                        'yes' if end_object else 'no'
                    )
                    if property_name in current_object['properties']:
                        log.warning(
                            '[%s:%i] Duplicated object property %s detected',
                            path, line_idx, property_name
                        )
                    current_object['properties'][property_name] = dict(
                        value=property_value,
                        comment=comment,
                        pre_comment_lines=next_property_pre_comment_lines
                    )
                    next_property_pre_comment_lines = []
                    if end_object:
                        add_object()
                        current_object = None
                    continue

            m = property_line.match(line)
            if m:
                # Object property line
                if not current_object:
                    log.warning(
                        '[%s:%i] Object property line detected, but no object currently opened',
                        path, line_idx)
                    continue
                property_name = m.group('name')
                line_return = m.group('value').endswith('\\')
                property_value = m.group('value').strip()
                if line_return:
                    property_value = property_value[0:-1].strip()
                end_object = m.group('end_mark')
                comment = m.group('comment')
                log.debug(
                    '[%s:%i] Property "%s" = "%s" (line return: %s, end object mark: %s)',
                    path, line_idx,
                    property_name, property_value,
                    'yes' if line_return else 'no',
                    'yes' if end_object else 'no'
                )
                assert not (line_return and (end_object or comment)), (
                    f'[{path}:{line_idx}] Can not have line return mark and end object mark '
                    'or comment on same line'
                )

                if property_name in current_object['properties']:
                    log.warning(
                        '[%s:%i] Duplicated object property %s detected',
                        path, line_idx, property_name
                    )

                if line_return:
                    property_before_line_return = dict(
                        name=property_name, value=property_value
                    )
                    continue

                current_object['properties'][property_name] = dict(
                    value=property_value,
                    comment=comment,
                    pre_comment_lines=next_property_pre_comment_lines
                )
                next_property_pre_comment_lines = []
                if end_object:
                    add_object()
                    current_object = None
                continue

            log.warning('[%s:%i] Unreconized line "%s"', path, line_idx, line)

    return objects


def parse_directory(path, objects_mappers=None, encoding=None):
    """ Parse icinga1 objects configuration directory """
    objects = dict(
        (obj_type, {} if info.get('unique', True) else [])
        for obj_type, info in map_obj_type_to_class.items()
    )
    objects_templates = dict(
        (obj_type, {})
        for obj_type in map_obj_type_to_class
    )
    disabled_objects = dict(
        (obj_type, {} if info.get('unique', True) else [])
        for obj_type, info in map_obj_type_to_class.items()
    )
    for root, dirs, files in os.walk(path):  # pylint: disable=unused-variable
        log.debug('Handle directory %s', root)
        for file in files:
            path = os.path.join(root, file)
            if not file.lower().endswith('.cfg'):
                log.debug('[%s] File ignored by name', path)
                continue
            log.debug('Handle file %s', path)
            file_objects = parse_file(
                path, encoding=encoding if encoding else default_encoding)

            for obj_type, objs in file_objects.items():
                for obj_info in objs:
                    if obj_type not in map_obj_type_to_class:
                        log.warning(
                            '[%s] Unknown/unsupported object type %s', path, obj_type)
                        continue
                    obj_class = map_obj_type_to_class[obj_type]['class']

                    try:
                        obj = obj_class(
                            filepath=obj_info['filepath'],
                            start_line=obj_info['start_line'],
                            end_line=obj_info['end_line'],
                            **obj_info['properties']
                        )
                    except Exception:
                        log.exception(
                            '[%s] Exception occured instanciating %s object '
                            '(from lines %d to %d):\n%s',
                            path, obj_type, obj_info['start_line'], obj_info['end_line'], obj_info)
                        raise

                    # Map object with registered objects mappers
                    for mapper in objects_mappers if objects_mappers else []:
                        mapper(obj)

                    # Handle registered objects
                    if obj.register:
                        if isinstance(objects[obj_type], dict):
                            if (
                                obj.name in objects[obj_type]
                                or obj.name in objects_templates[obj_type]
                                or obj.name in disabled_objects[obj_type]
                            ):
                                log.fatal(
                                    '[%s] Duplicated object %s with name %s found',
                                    path, obj_type, obj.name)
                                sys.exit(1)
                            objects[obj_type][obj.name] = obj
                        else:
                            objects[obj_type].append(obj)
                        log.debug('New object %s with name %s found:\n%s',
                                  obj_type, obj.name, obj.format())

                    # Handle templates
                    if obj.template_name and obj.template_name in objects_templates[obj_type]:
                        log.fatal(
                            '[%s] Duplicated object %s template with name %s found:\n%s\n%s',
                            path,
                            obj_type,
                            obj.template_name,
                            objects_templates[obj_type][obj.template_name].format(),
                            obj.format()
                        )
                        sys.exit(1)
                    elif not obj.register and not obj.template_name:
                        log.debug(
                            "Not registered object %s without template name: "
                            "keep it as disabled\n%s",
                            obj_type, obj.format())
                        if isinstance(disabled_objects[obj_type], dict):
                            if (
                                obj.name in objects[obj_type]
                                or obj.name in disabled_objects[obj_type]
                                or obj.name in objects_templates[obj_type]
                            ):
                                log.fatal(
                                    '[%s] Duplicated object %s with name %s found',
                                    path, obj_type, obj.name)
                                sys.exit(1)
                            disabled_objects[obj_type][obj.name] = obj
                        else:
                            disabled_objects[obj_type].append(obj)
                    elif obj.template_name:
                        log.debug(
                            'New object %s template %s found:\n%s',
                            obj_type, obj.template_name, obj.format())
                        objects_templates[obj_type][obj.template_name] = obj

    # Handle template inheritance
    for obj_type in objects_templates:
        for name in objects_templates[obj_type]:
            objects_templates[obj_type][name].set_template(objects_templates[obj_type])

    # Set objects templates
    for obj_type in objects:
        if isinstance(objects[obj_type], dict):
            for name in objects[obj_type]:
                objects[obj_type][name].set_template(objects_templates[obj_type])
        else:
            for obj in objects[obj_type]:
                obj.set_template(objects_templates[obj_type])

    # Set disabled objects templates
    for obj_type in disabled_objects:
        if isinstance(disabled_objects[obj_type], dict):
            for name in disabled_objects[obj_type]:
                disabled_objects[obj_type][name].set_template(objects_templates[obj_type])
        else:
            for obj in disabled_objects[obj_type]:
                obj.set_template(objects_templates[obj_type])

    return objects, objects_templates, disabled_objects


def format_object(obj):
    lines = [f'Object {obj["type"]}:']
    for name, info in obj['properties'].items():
        if info['pre_comment_lines']:
            for line in info['pre_comment_lines']:
                lines.append(f'    # {line}')
        lines.append(
            f'    - {name}: "{info["value"]}"'
            f'{(" # " + info["comment"]) if info["comment"] else ""}'
        )
    return "\n".join(lines)
