""" Icinga2 export helpers """

from copy import copy
import logging
import re

log = logging.getLogger(__name__)

# Store Icinga1 mapping to Icinga2
icinga1_object_mapping = {}
icinga1_object_displayname_mapping = False
icinga1_excluded_objects = {}
icinga1_custom_macro_mapping = {}
icinga1_custom_user_macro_values_mapping = {}
icinga1_ignored_unsupported_properties = []
icinga1_host_parents_mapping_policies = ['individual-dependency', 'variable']
icinga1_host_parents_mapping_policy = icinga1_host_parents_mapping_policies[0]
icinga1_interval_length = 60
var_name_lowercase = False


def handle_options(options):
    global var_name_lowercase  # pylint: disable=global-statement
    global icinga1_object_displayname_mapping  # pylint: disable=global-statement
    global icinga1_host_parents_mapping_policy  # pylint: disable=global-statement
    global icinga1_interval_length  # pylint: disable=global-statement

    # Handle objects names mapping
    for mapping in options.map_objects:
        if mapping['obj_type'] not in icinga1_object_mapping:
            icinga1_object_mapping[mapping['obj_type']] = {}
        else:
            assert mapping['icinga1_name'] not in icinga1_object_mapping[mapping['obj_type']], (
                f'Duplicated mapping specified for {mapping["obj_type"]} object '
                f'{mapping["icinga1_name"]}')
        icinga1_object_mapping[mapping['obj_type']][mapping['icinga1_name']] = \
            mapping['icinga2_name']

    # Handle excluded objects
    for exclusion in options.excluded_objects:
        icinga1_name = exclusion['icinga1_name'].lower().strip()
        if exclusion['obj_type'] not in icinga1_excluded_objects:
            icinga1_excluded_objects[exclusion['obj_type']] = [icinga1_name]
        elif icinga1_name not in icinga1_excluded_objects[exclusion['obj_type']]:
            icinga1_excluded_objects[exclusion['obj_type']].append(icinga1_name)

    # Handle custom macro names mapping
    for opt in options.map_custom_macros:
        mapping = [part.strip() for part in opt.split('=')]
        assert len(mapping) == 2, (
            f'Invalid --map-custom-macro value "{opt}": must be in format '
            '--map-custom-macro icinga1_custom_macro=icinga2_var')
        icinga1_custom_macro_mapping[mapping[0]] = mapping[1]

    # Handle custom user macro values mapping
    for opt in options.map_custom_user_macro_values:
        mapping = [part.strip() for part in opt.split('=')]
        assert len(mapping) == 2, (
            f'Invalid --map-custom-user-macro-value value "{opt}": must be in format '
            '--map-custom-user-macro-value USERx=icinga2_value')
        icinga1_custom_user_macro_values_mapping[mapping[0]] = mapping[1]

    if options.var_name_lowercase:
        var_name_lowercase = True

    if options.ignore_unsupported_properties:
        icinga1_ignored_unsupported_properties.extend([
            prop_name.lower().strip()
            for prop_name in options.ignore_unsupported_properties
        ])

    if options.host_parents_mapping_policy:
        icinga1_host_parents_mapping_policy = options.host_parents_mapping_policy

    if options.map_objects_display_name:
        icinga1_object_displayname_mapping = True

    if options.icinga1_interval_length:
        icinga1_interval_length = options.icinga1_interval_length


def map_to_icinga1_object_type(obj_type):
    mapping = {
        'user': 'contact',
        'usergroup': 'contactgroup',
        'notificationcommand': 'command',
        'checkcommand': 'command',
        'eventcommand': 'command',
        'period': 'timeperiod',
    }
    return mapping.get(obj_type.lower(), obj_type.lower())


def map_icinga1_object(obj_type, name):
    if not isinstance(obj_type, str):
        log.warning(
            'map_icinga1_object(%s, %s): obj_type is not a string (%s)',
            obj_type, name, type(obj_type))
        return name
    mapped = icinga1_object_mapping.get(map_to_icinga1_object_type(obj_type), {}).get(name, name)
    log.debug('map_icinga1_object(%s, %s): %s', obj_type, name, mapped)
    return mapped


def map_icinga1_object_display_name(obj_type, name):
    if icinga1_object_displayname_mapping:
        return map_icinga1_object(obj_type, name)
    return name


def escape_string(value):
    # First protect runtime macro by replacing their "$" with "___MACRO___"
    for item in re.finditer(r'\$([a-zA-Z0-9\.\_]+)\$', value):
        value = value.replace(
            item.group(0),
            f'___MACRO___{item.group(1)}___MACRO___'
        )

    # Escape special characters
    # pylint: disable=line-too-long
    # https://icinga.com/docs/icinga2/latest/doc/17-language-reference/#string-literals-escape-sequences
    value = value.replace('\\', '\\\\')
    value = value.replace('"', '\\"')
    value = value.replace('\t', '\\t')
    value = value.replace('\r', '\\r')
    value = value.replace('\n', '\\n')

    # Escape "$"
    value = value.replace('$', '$$')

    # Restore "$" of legitimate runtime macro
    value = value.replace('___MACRO___', '$')
    return value


def map_icinga1_macro(value, map_command_args=False, map_command_type=None):
    assert not map_command_args or map_command_type, \
        "if map_command_args == True, you must also provide map_command_type"
    mapping = {
        'CONTACTNAME': 'user.name',
        'CONTACTALIAS': 'user.display_name',
        'CONTACTEMAIL': 'user.email',
        'CONTACTPAGER': 'user.pager',
        'SERVICEDESC': 'service.name',
        'SERVICEDISPLAYNAME': 'service.display_name',
        'SERVICECHECKCOMMAND': 'service.check_command',
        'SERVICESTATE': 'service.state',
        'SERVICESTATEID': 'service.state_id',
        'SERVICESTATETYPE': 'service.state_type',
        'SERVICEATTEMPT': 'service.check_attempt',
        'MAXSERVICEATTEMPT': 'service.max_check_attempts',
        'LASTSERVICESTATE': 'service.last_state',
        'LASTSERVICESTATEID': 'service.last_state_id',
        'LASTSERVICESTATETYPE': 'service.last_state_type',
        'LASTSERVICESTATECHANGE': 'service.last_state_change',
        'SERVICEDOWNTIME': 'service.downtime_depth',
        'SERVICEDURATIONSEC': 'service.duration_sec',
        'SERVICELATENCY': 'service.latency',
        'SERVICEEXECUTIONTIME': 'service.execution_time',
        'SERVICEOUTPUT': 'service.output',
        'SERVICEPERFDATA': 'service.perfdata',
        'LASTSERVICECHECK': 'service.last_check',
        'SERVICENOTES': 'service.notes',
        'SERVICENOTESURL': 'service.notes_url',
        'SERVICEACTIONURL': 'service.action_url',
        'HOSTNAME': 'host.name',
        'HOSTADDRESS': 'host.address',
        'HOSTADDRESS6': 'host.address6',
        'HOSTDISPLAYNAME': 'host.display_name',
        'HOSTALIAS': 'host.display_name',
        'HOSTCHECKCOMMAND': 'host.check_command',
        'HOSTSTATE': 'host.state',
        'HOSTSTATEID': 'host.state_id',
        'HOSTSTATETYPE': 'host.state_type',
        'HOSTATTEMPT': 'host.check_attempt',
        'MAXHOSTATTEMPT': 'host.max_check_attempts',
        'LASTHOSTSTATE': 'host.last_state',
        'LASTHOSTSTATEID': 'host.last_state_id',
        'LASTHOSTSTATETYPE': 'host.last_state_type',
        'LASTHOSTSTATECHANGE': 'host.last_state_change',
        'HOSTDOWNTIME': 'host.downtime_depth',
        'HOSTDURATIONSEC': 'host.duration_sec',
        'HOSTLATENCY': 'host.latency',
        'HOSTEXECUTIONTIME': 'host.execution_time',
        'HOSTOUTPUT': 'host.output',
        'HOSTPERFDATA': 'host.perfdata',
        'LASTHOSTCHECK': 'host.last_check',
        'HOSTNOTES': 'host.notes',
        'HOSTNOTESURL': 'host.notes_url',
        'HOSTACTIONURL': 'host.action_url',
        'TOTALSERVICES': 'host.num_services',
        'TOTALSERVICESOK': 'host.num_services_ok',
        'TOTALSERVICESWARNING': 'host.num_services_warning',
        'TOTALSERVICESUNKNOWN': 'host.num_services_unknown',
        'TOTALSERVICESCRITICAL': 'host.num_services_critical',
        'COMMANDNAME': 'command.name',
        'NOTIFICATIONTYPE': 'notification.type',
        'NOTIFICATIONAUTHOR': 'notification.author',
        'NOTIFICATIONCOMMENT': 'notification.comment',
        'NOTIFICATIONAUTHORNAME': 'notification.author',
        'NOTIFICATIONAUTHORALIAS': 'notification.author',
        'TIMET': 'icinga.timet',
        'LONGDATETIME': 'icinga.long_date_time',
        'SHORTDATETIME': 'icinga.short_date_time',
        'DATE': 'icinga.date',
        'TIME': 'icinga.time',
        'PROCESSSTARTTIME': 'icinga.uptime',
        'TOTALHOSTSUP': 'icinga.num_hosts_up',
        'TOTALHOSTSDOWN': 'icinga.num_hosts_down',
        'TOTALHOSTSUNREACHABLE': 'icinga.num_hosts_unreachable',
        'TOTALHOSTPROBLEMS': 'icinga.num_hosts_down',
    }

    unmappable = {
        'TOTALHOSTPROBLEMSUNHANDLED': (
            'use instead: (icinga.num_hosts_down + icinga.num_hosts_unreachable - '
            'icinga.num_hosts_in_downtime - icinga.num_hosts_acknowledged)'),
        'TOTALSERVICEPROBLEMS': (
            'use instead: (icinga.num_services_warning + icinga.num_services_critical '
            '+ icinga.num_services_unknown)'),
        'TOTALSERVICEPROBLEMSUNHANDLED': (
            'use instead: (icinga.num_services_warning + icinga.num_services_critical + '
            'icinga.num_services_unknown - icinga.num_hosts_in_downtime - '
            'icinga.num_hosts_acknowledged)'),
        'TOTALSERVICESWARNINGUNHANDLED': None,
        'TOTALSERVICESCRITICALUNHANDLED': None,
        'TOTALSERVICESUNKNOWNUNHANDLED': None,
        'TOTALHOSTSDOWNUNHANDLED': None,
        'TOTALHOSTSUNREACHABLEUNHANDLED': None,
    }

    is_command_arg = re.compile(r'^ARG[0-9]+$')
    orig_value = copy(value)
    # pylint: disable=line-too-long
    for item in re.finditer(r'\$(?P<varname>((USER(?P<custom_user_macro_idx>[0-9]+)|(_(?P<custom_macro_type>HOST|SERVICE|CONTACT))?(?P<custom_macro_name>[^\$]+))))\$', value):
        log.debug(
            'map_icinga1_macro(%s): detect macro "%s" (varname="%s")',
            orig_value, item.group(0), item.group('varname')
        )
        if item.group('custom_user_macro_idx'):
            # Custom user macros
            log.debug(
                'map_icinga1_macro(%s): detect custom user macro "%s"',
                orig_value, item.group('varname'),
            )
            if item.group('varname') in icinga1_custom_user_macro_values_mapping:
                log.debug(
                    'map_icinga1_macro(%s): map custom user macro "%s" value to "%s"',
                    orig_value, item.group('varname'),
                    icinga1_custom_user_macro_values_mapping[item.group('varname')]
                )
                value = value.replace(
                    item.group(0),
                    icinga1_custom_user_macro_values_mapping[item.group('varname')]
                )
            else:
                log.warning(
                    'No mapping exist for custom user macro "%s" in string "%s". '
                    'You could map it using --map-custom-user-macro-value parameter.',
                    item.group('varname'), orig_value
                )
        elif item.group('custom_macro_type'):
            # Custom macros
            log.debug(
                'map_icinga1_macro(%s): detect custom %s macro with name "%s"',
                orig_value, item.group('custom_macro_type'), item.group('custom_macro_name')
            )
            icinga2_var_name = map_icinga1_custom_variable_name(
                item.group('custom_macro_name'))
            log.debug(
                'map_icinga1_macro(%s): map custom %s macro "%s" to "$%s$"',
                orig_value, item.group('custom_macro_type'),
                item.group(0), icinga2_var_name
            )
            value = value.replace(
                item.group(0),
                f'${icinga2_var_name}$'
            )
        elif item.group('varname') in mapping:
            # Mappable macros
            log.debug(
                'map_icinga1_macro(%s): map macro "%s" to "$%s$"',
                orig_value, item.group(0), mapping[item.group('varname')]
            )
            value = value.replace(
                item.group(0),
                f'${mapping[item.group("varname")]}$'
            )
        elif item.group('varname') in unmappable:
            # Unmappable macros
            log.warning(
                'Unmappable runtime macro "%s" detected in the following string:\n'
                '  "%s"\nSolution: %s',
                item.group(0), orig_value,
                (
                    unmappable[item.group('varname')] if unmappable[item.group('varname')]
                    else "no equivalent runtime macro exist in Icinga2 :("
                )
            )
        elif map_command_args and is_command_arg.match(item.group('varname')):
            # Command args macros
            icinga2_var_name = f'{map_command_type.upper()}_{item.group("varname")}'
            if var_name_lowercase:
                icinga2_var_name = icinga2_var_name.lower()
            log.debug(
                'map_icinga1_macro(%s): command argument macro "%s" map to "$%s$"',
                orig_value, item.group(0), icinga2_var_name
            )
            value = value.replace(
                item.group(0),
                f'${icinga2_var_name}$'
            )
        elif var_name_lowercase:
            log.debug(
                'map_icinga1_macro(%s): no icinga2 mapping for macro "%s" but put it in lowercase',
                orig_value, item.group(0)
            )
            value = value.replace(
                item.group(0),
                item.group(0).lower()
            )
        else:
            log.debug(
                'map_icinga1_macro(%s): no icinga2 mapping for macro "%s"',
                orig_value, item.group(0)
            )
    return value


def map_icinga1_custom_variable_name(var_name):
    if var_name in icinga1_custom_macro_mapping:
        icinga2_var_name = icinga1_custom_macro_mapping[var_name]
    else:
        icinga2_var_name = var_name
        if var_name_lowercase:
            icinga2_var_name = icinga2_var_name.lower()
    log.debug(
        'map_icinga1_custom_variable_name(%s): map to "%s"',
        var_name, icinga2_var_name
    )
    return icinga2_var_name


def is_ignored_icinga1_unsupported_property(name):
    return name.lower().strip() in icinga1_ignored_unsupported_properties


def is_icinga1_excluded_object(obj_type, name):
    return name and (
        name.lower().strip() in icinga1_excluded_objects.get(
            map_to_icinga1_object_type(obj_type), []
        )
    )


def export(config, fd, disabled=False):
    config = config if isinstance(config, list) else [config]
    lines = []
    for line in config:
        lines.extend(line.split("\n"))
    if disabled:
        lines = [f'# {line}' for line in lines]
    return fd.write("\n".join(lines))


def export_simple_mapping_properties(obj_type, obj, variables_supported=False, prefix=None):
    lines = []
    prefix = prefix if prefix else ""

    # Handle "simple mapping" directives
    for name, prop in obj.properties.items():  # pylint: disable=unused-variable
        if not prop.icinga2_simple_mapping or prop.is_obsolete_in_icinga2():
            continue
        if not prop.is_defined() and not prop.always_exported:
            continue
        if (
            not prop.is_supported_by_icinga2()
            and not prop.is_ignored_icinga1_unsupported_property()
        ):
            log.warning(
                '%s: Unsupported Icinga1 %s property (value = %s)',
                obj, name, prop.get_value())
        directive = prop.get_icinga2_directive(
            prefix=(
                prefix if prop.is_supported_by_icinga2()
                else f'{prefix}# Unsupported Icinga1 '
            ),
            object_type=obj_type
        )
        if directive:
            lines.append(directive)

    # Handle unknown properties
    for name, prop in obj.unknown_properties.items():
        lines.append(
            prop.get_icinga2_directive(
                prefix=f'{prefix}# Unknown Icinga1 property ',
                object_type=obj_type
            )
        )

    # Handle variables
    if obj.variables:
        if not variables_supported:
            log.warning(
                '%s: custom variables could not be keeped in Icinga2 (unsupported for %s)',
                obj, obj_type)
            prefix = f'{prefix}# Unsupported Icinga1 variable '
        for name, var in obj.variables.items():
            lines.append(
                var.get_icinga2_directive(prefix=prefix, object_type=obj_type)
            )

    return lines


def export_simple_object_type(obj_type, obj, fd, variables_supported=False, apply=False,
                              disabled=False):
    # Handle other "simple mapping" directives
    lines = export_simple_mapping_properties(
        obj_type, obj,
        prefix='  ', variables_supported=variables_supported
    )
    return export_object_config_lines(obj_type, obj, fd, lines, apply=apply, disabled=disabled)


def export_object_config_lines(obj_type, obj, fd, lines, apply=False, disabled=False):
    config = '%s %s "%s" {\n%s\n}\n\n' % (
        ("apply" if apply else "object") if obj.register or not obj.template_name else "template",
        obj_type,
        map_icinga1_object(
            obj_type, obj.name if obj.register or not obj.template_name else obj.template_name),
        "\n".join(lines)
    )
    if obj.register and obj.template_name and obj.template_name != obj.name:
        log.debug(
            "%s %s is registered as an object but have a template name that defer from "
            "object name (%s != %s): Also register it as template with name %s",
            obj_type, obj, obj.template_name, obj.name, obj.template_name
        )
        config += 'template %s "%s" {\n%s\n}\n\n' % (
            obj_type,
            map_icinga1_object(obj_type, obj.template_name),
            "\n".join(lines)
        )
    log.debug("%s %s output: %s", obj_type, obj, config)
    return export(config, fd, disabled=disabled)


def export_host_and_service_notifications_variable(obj_type, obj):
    # Handle contacts, contact_groups, notification_interval,
    # notification_period, notifications_options and first_notification_delay:
    # Convert it as object notifications variable
    # vars.notifications = {
    #   users = []          # From contacts
    #   ignored_users = []
    #   groups = []         # From contact_groups
    #   ignored_groups = []
    #   interval = Xm       # From notification_interval
    #   delay = Xm          # From first_notification_delay
    #   period = "X"        # From notification_period
    #   states = []         # From notification_options
    #   types = []
    # }
    # See:
    # pylint: disable=line-too-long
    #  https://icinga.com/docs/icinga2/latest/doc/23-migrating-from-icinga-1x/#manual-config-migration-hints-for-notifications
    #  https://icinga.com/docs/icinga2/latest/doc/03-monitoring-basics/#notifications-users-from-hostservice

    # If no of this directive is present, return nothing
    present = False
    for property_name in [
        'contacts', 'contact_groups', 'notification_interval', 'notification_period',
        'notification_options', 'first_notification_delay'
    ]:
        if obj.is_defined(property_name):
            present = True
            break
    if not present:
        return []

    # Otherwise
    lines = [
        '',
        '  # Convert Icinga1 contacts, contact_group, notification_interval,',
        '  # notification_period, notifications_options and first_notification_delay',
        '  # as notifications variable',
    ]
    lines.append('  vars.notifications = {')
    if obj.contacts:
        lines.append(obj.properties['contacts'].get_icinga2_directive(
            prefix='    # ', object_type=obj_type))
        included = []
        excluded = []
        for contact in obj.contacts:
            if contact.startswith('!'):
                excluded.append(contact[1:])
            else:
                included.append(contact)
        if included:
            lines.append(
                # pylint: disable=consider-using-f-string
                '    users = ["%s"]' % '", "'.join([
                    map_icinga1_object('contact', contact) for contact in included
                ])
            )
        if excluded:
            lines.append(
                '    # Icinga2 does not permit to ignore some users, but Icinga1 ignored users '
                'list is keep here')
            lines.append(
                # pylint: disable=consider-using-f-string
                '    ignored_users = ["%s"]' % '", "'.join([
                    map_icinga1_object('contact', contact) for contact in excluded
                ])
            )

    if obj.contact_groups:
        lines.append(obj.properties['contact_groups'].get_icinga2_directive(
            prefix='    # ', object_type=obj_type))
        included = []
        excluded = []
        for group in obj.contact_groups:
            if group.startswith('!'):
                excluded.append(group[1:])
            else:
                included.append(group)
        if included:
            lines.append(
                # pylint: disable=consider-using-f-string
                '    groups = ["%s"]' % '", "'.join([
                    map_icinga1_object('contactgroup', group) for group in included
                ])
            )
        if excluded:
            lines.append(
                '    # Icinga2 does not permit to ignore some groups, but Icinga1 ignored groups '
                'list is keep here')
            lines.append(
                # pylint: disable=consider-using-f-string
                '    ignored_groups = ["%s"]' % '", "'.join([
                    map_icinga1_object('contactgroup', group) for group in excluded
                ])
            )

    for property_name in ['notification_interval', 'first_notification_delay',
                          'notification_period', 'notification_options']:
        if not obj.is_defined(property_name):
            continue
        lines.append(
            obj.properties[property_name].get_icinga2_directive(
                prefix='    ', object_type=obj_type)
        )

    lines.append('  }')
    return lines


def export_host(host, fd, disabled=False):
    lines = []

    # Handle alias / display_name properties
    if host.is_defined('alias') and host.is_defined('display_name'):
        log.warning(
            "Host %s have alias and display_name directive, but Icinga2 only handle one "
            "display_name parameter. Keep display_name attribute and put alias value as comment",
            host)
        if host.properties['alias'].comment:
            lines.append(f'  # {host.properties["alias"].comment}')
        if host.properties['display_name'].comment:
            lines.append(f'  # {host.properties["display_name"].comment}')
        lines.append(
            f'  display_name = {host.properties["display_name"].get_icinga2_value()}'
            f'  # Icinga1 alias = {host.properties["alias"].get_icinga2_value()}'
        )
    elif host.alias:
        lines.append(
            host.properties['alias'].get_icinga2_directive(
                prefix='  ', object_type='Host')
        )
    elif host.display_name:
        lines.append(
            host.properties['display_name'].get_icinga2_directive(
                prefix='  ', object_type='Host')
        )

    if icinga1_host_parents_mapping_policy == 'variable' and host.parents:
        host.add_variable(
            'parents', value=','.join(host.parents), array=True,
            object_mapping_type='host')

    # Handle other "simple mapping" directives
    lines += export_simple_mapping_properties(
        'Host', host,
        variables_supported=True, prefix='  '
    )

    # Handle host without check_command (required since Icinga2)
    if not host.is_defined('check_command', use_template=True) and not host.use:
        lines += [
            '',
            '  # Host have no check_command in Icinga1 but this property is required',
            '  # since Icinga2: use "dummy" CheckCommand',
            '  check_command = "dummy"'
        ]

    lines += export_host_and_service_notifications_variable('Host', host)

    if not export_object_config_lines("Host", host, fd, lines, disabled=disabled):
        return False

    if icinga1_host_parents_mapping_policy == 'individual-dependency':
        # Handle parents as Host-to-Host dependencies
        # pylint: disable=line-too-long
        # https://icinga.com/docs/icinga2/latest/doc/23-migrating-from-icinga-1x/#manual-config-migration-hints-for-host-parents
        host_name = map_icinga1_object("host", host.name)
        for parent in host.parents:
            parent = map_icinga1_object("host", parent)
            config = [
                f'object Dependency "{parent}-parent-of-{host_name}" ' + "{",
                f'    parent_host_name = "{parent}"',
                f'    child_host_name = "{host_name}"',
                '}\n\n'
            ]
            if not export(config, fd, disabled=disabled):
                return False
    return True


def export_hostgroup(hostgroup, fd, disabled=False):
    return export_simple_object_type(
        'HostGroup', hostgroup,
        fd, disabled=disabled
    )


def export_hostdependency(hostdependency, fd, disabled=False):
    if not hostdependency.register and hostdependency.template_name:
        log.warning(
            'Template of Icinga1 hostdependency is not supported, ignore the following one: %s',
            hostdependency.format())
        return False

    host_names = hostdependency.properties['host_name'].get_value()
    if len(host_names) == 1:
        parent = map_icinga1_object("host", host_names[0])
        config = [
            f'apply Dependency "{hostdependency.name}" to Host ' + "{",
            f'  parent_host_name = "{escape_string(parent)}"'
        ]
    elif host_names:
        parent = map_icinga1_object(
            "host", hostdependency.properties["host_name"].get_icinga2_value())
        config = [
            f'apply Dependency "{hostdependency.name}" for (parent in {parent}) to Host ' + '{',
            '  parent_host_name = parent'
        ]
    else:
        log.warning(
            'Icinga1 hostdependency without host_name is not supported, '
            'ignore the following one: %s', hostdependency.format())
        return False

    config += export_simple_mapping_properties(
        'Dependency', hostdependency,
        prefix='  ', variables_supported=False
    )
    config += ["}\n\n"]

    log.debug("Dependency %s output: %s", hostdependency, config)
    return export(config, fd, disabled=disabled)


def export_service(service, fd, disabled=False):
    lines = export_simple_mapping_properties(
        'Service', service,
        variables_supported=True, prefix='  '
    )
    lines += export_host_and_service_notifications_variable('Service', service)
    return export_object_config_lines(
        "Service", service,
        fd, lines, apply=True, disabled=disabled
    )


def export_servicegroup(servicegroup, fd, disabled=False):
    return export_simple_object_type(
        'ServiceGroup', servicegroup,
        fd, variables_supported=False, disabled=disabled
    )


def export_servicedependency(servicedependency, fd, disabled=False):
    if not servicedependency.register and servicedependency.template_name:
        log.warning(
            'Template of Icinga1 servicedependency is not supported, ignore the following one: %s',
            servicedependency.format())
        return False

    host_names = servicedependency.properties['host_name'].get_value()
    if len(host_names) == 1:
        parent = map_icinga1_object("host", host_names[0])
        config = [
            f'apply Dependency "{servicedependency.name}" to Service ' + "{",
            f'  parent_host_name = "{escape_string(parent)}"'
        ]
    elif host_names:
        parent = map_icinga1_object(
            "host", servicedependency.properties["host_name"].get_icinga2_value())
        config = [
            f'apply Dependency "{servicedependency.name}" for (parent in {parent}) to Service '
            '{', '  parent_host_name = parent'
        ]
    else:
        log.warning(
            'Icinga1 servicedependency without host_name is not supported, '
            'ignore the following one: %s', servicedependency.format())
        return False

    for prop_name in ['dependent_host_name', 'dependent_hostgroup_name']:
        if servicedependency.dependent_service_description:
            directive = servicedependency.properties[prop_name].get_icinga2_directive(
                prefix='  ', additional_clause=(
                    'service.name == '
                    f'"{escape_string(servicedependency.dependent_service_description)}"'
                )
            )
            if directive:
                config.append(directive)
        else:
            servicedependency.properties[prop_name].icinga2_simple_mapping = True

    config += export_simple_mapping_properties(
        'Dependency', servicedependency,
        prefix='  ', variables_supported=False
    )
    config += ["}\n\n"]

    log.debug("Dependency %s output: %s", servicedependency, config)
    return export(config, fd, disabled=disabled)


def export_contact(contact, fd, disabled=False):
    lines = export_simple_mapping_properties(
        'User', contact,
        prefix='  ', variables_supported=True
    )

    # Check service_notifications_enabled == host_notifications_enabled
    if contact.service_notifications_enabled != contact.host_notifications_enabled:
        lines += [
            '  # Icinga1 service_notifications_enabled != host_notifications_enabled',
            '  # Icinga2 have only one enable_notifications parameter: '
            'keep host_notifications_enabled value',
            contact.properties['service_notifications_enabled'].get_icinga2_directive(
                prefix='  # ', object_type='User')
        ]

    # Check service_notification_period == host_notification_period
    if contact.service_notification_period != contact.service_notification_period:
        lines += [
            '',
            '  # Icinga1 service_notification_period != host_notification_period',
            '  # Icinga2 have only one period parameter: keep host_notification_period value',
            contact.properties['service_notification_period'].get_icinga2_directive(
                prefix='  # ', object_type='User')
        ]

    # Handle host_notification_options & service_notification_options conversion as states & types
    states = contact.properties['host_notification_options'].get_icinga2_states()
    states += contact.properties['service_notification_options'].get_icinga2_states()
    # Deduplicate
    states = list(dict.fromkeys(states))

    types = contact.properties['host_notification_options'].get_icinga2_types()
    types += contact.properties['service_notification_options'].get_icinga2_types()
    # Deduplicate
    types = list(dict.fromkeys(types))

    lines += [
        '',
        '  # Icinga1 host_notification_options & service_notification_options converted '
        'as Icinga2 states & types',
        contact.properties['host_notification_options'].get_icinga2_directive(
            prefix='  # ', object_type='User', raw_values=True),
        contact.properties['service_notification_options'].get_icinga2_directive(
            prefix='  # ', object_type='User', raw_values=True),
        f'  states = [{", ".join(states)}]',
        f'  types = [{", ".join(types)}]',
    ]

    return export_object_config_lines('User', contact, fd, lines, disabled=disabled)


def export_contactgroup(contactgroup, fd, disabled=False):
    return export_simple_object_type(
        'UserGroup', contactgroup,
        fd, variables_supported=False, disabled=disabled
    )


def detect_command_types(command, icinga1_objects, icinga1_templates):
    # Try to detect command types by looking what objects are using it
    command_types = []

    # Hosts
    for host_name, host in icinga1_objects['host'].items():
        if host.properties['check_command'].command == command.name:
            log.debug(
                'detect_command_type(%s): Host %s use it as check_command',
                command.name, host_name
            )
            if 'CheckCommand' not in command_types:
                command_types.append("CheckCommand")
        if host.properties['event_handler'].command == command.name:
            log.debug(
                'detect_command_type(%s): Host %s use it as event_handler',
                command.name, host_name
            )
            if 'EventCommand' not in command_types:
                command_types.append("EventCommand")

    # Host templates
    for tpl_name, tpl in icinga1_templates['host'].items():
        if tpl.properties['check_command'].command == command.name:
            log.debug(
                'detect_command_type(%s): Host template %s use it as check_command',
                command.name, tpl_name
            )
            if 'CheckCommand' not in command_types:
                command_types.append("CheckCommand")
        if tpl.properties['event_handler'].command == command.name:
            log.debug(
                'detect_command_type(%s): Host template %s use it as event_handler',
                command.name, tpl_name
            )
            if 'EventCommand' not in command_types:
                command_types.append("EventCommand")

    # Services
    for service in icinga1_objects['service']:
        if service.properties['check_command'].command == command.name:
            log.debug(
                'detect_command_type(%s): %s use it as check_command',
                command.name, service
            )
            if 'CheckCommand' not in command_types:
                command_types.append("CheckCommand")
        if service.properties['event_handler'].command == command.name:
            log.debug(
                'detect_command_type(%s): %s use it as event_handler',
                command.name, service
            )
            if 'EventCommand' not in command_types:
                command_types.append("EventCommand")

    # Service templates
    for tpl_name, tpl in icinga1_templates['service'].items():
        if tpl.properties['check_command'].command == command.name:
            log.debug(
                'detect_command_type(%s): Service template %s use it as check_command',
                command.name, tpl_name
            )
            if 'CheckCommand' not in command_types:
                command_types.append("CheckCommand")
        if tpl.properties['event_handler'].command == command.name:
            log.debug(
                'detect_command_type(%s): Service template %s use it as event_handler',
                command.name, tpl_name
            )
            if 'EventCommand' not in command_types:
                command_types.append("EventCommand")

    # Contacts
    for contact_name, contact in icinga1_objects['contact'].items():
        for property_name in ['host_notification_commands', 'service_notification_commands']:
            if (
                contact.is_defined(property_name)
                and command.name in getattr(contact, property_name)
            ):
                log.debug(
                    'detect_command_type(%s): Contact %s use it as %s',
                    command.name, contact_name, property_name
                )
                command_types.append("NotificationCommand")
                break

    # Contact templates
    if "NotificationCommand" not in command_types:
        for tpl_name, tpl in icinga1_templates['contact'].items():
            for property_name in ['host_notification_commands', 'service_notification_commands']:
                if tpl.is_defined(property_name) and command.name in getattr(tpl, property_name):
                    log.debug(
                        'detect_command_type(%s): Contact template %s use it as %s',
                        command.name, tpl_name, property_name
                    )
                    command_types.append("NotificationCommand")
                break

    if not command_types:
        log.debug(
            'detect_command_type(%s): command not detected as used. Keep it as CheckCommand only.',
            command.name
        )
        return ['CheckCommand']
    log.debug('detect_command_type(%s): command detected as used as %s',
              command.name, ', '.join(command_types))
    return command_types


def export_command(command, command_types, fd, disabled=False):
    for command_type in command_types:
        export_simple_object_type(
            command_type, command,
            fd, variables_supported=False, disabled=disabled
        )


def export_timeperiod(timeperiod, fd, disabled=False):
    # Handle "simple mapping" directives
    lines = export_simple_mapping_properties(
        'TimePeriod', timeperiod,
        variables_supported=True, prefix='  '
    )

    # Handles week days to ranges
    ranges = []
    for name, prop in timeperiod.properties.items():  # pylint: disable=unused-variable
        if type(prop).__name__ != 'TimePeriodWeekDayProperty':
            continue
        if not prop.is_defined():
            continue
        ranges.append(
            prop.get_icinga2_directive(prefix="    ", object_type='Period')
        )
    if ranges:
        lines.append('\n  ranges = {\n%s\n  }' % '\n'.join(ranges))

    return export_object_config_lines("TimePeriod", timeperiod, fd, lines, disabled=disabled)
