""" Icinga objects """

import logging
import re

from .exceptions import (
    IcingaObjectUnknownProperty, IcingaObjectWithoutName, IcingaObjectUnknownTemplate)
from .icinga2 import map_icinga1_object, map_icinga1_custom_variable_name
from .icinga2 import is_icinga1_excluded_object
from .properties import (
    StringProperty,
    UseProperty,
    AddressProperty,
    IntegerProperty,
    FloatProperty,
    BooleanProperty,
    CommandProperty,
    CommandLineProperty,
    CommaSeparatedListProperty,
    HostMembersProperty,
    ContactMembersProperty,
    ObjectServiceGroupMembersProperty,
    HostGroupMembersProperty,
    ServiceGroupMembersProperty,
    ContactGroupMembersProperty,
    HostNotificationOptionsProperty,
    ServiceNotificationOptionsProperty,
    HostFailureCriteriaProperty,
    ServiceFailureCriteriaProperty,
    TimeUnitProperty,
    TimePeriodWeekDayProperty,
    IcingaObjectArrayVariable,
    IcingaObjectVariable,
    InheritsParentProperty,
    ServiceDescriptionProperty,
    NotificationFailureCriteriaProperty,
    DisplayNameProperty
)


log = logging.getLogger(__name__)


class IcingaObject:

    type = None
    name_property = None
    templates = []
    variables = {}

    properties = {}
    unknown_properties = {}
    properties_alias = {}
    _properties = {}

    def __init__(self, filepath=None, start_line=None, end_line=None, **properties):
        self.filepath = filepath
        self.start_line = start_line
        self.end_line = end_line
        # Initialize properties
        self.init_properties()
        for name, info in properties.items():
            if not self.set_property(name, **info):
                log.warning(
                    '%s: unknown or unsupported property "%s" = "%s"',
                    self.type,
                    name,
                    info['value']
                )
                self.unknown_properties[name] = StringProperty(
                    name, self
                )
                self.unknown_properties[name].set_info(**info)
        if self.name_property and self.register and not self.is_defined(self.name_property):
            raise IcingaObjectWithoutName(self)

    def init_properties(self):
        self.properties = {}
        self.properties['use'] = UseProperty('use', self)
        # template_name == directive name in Icinga1 config, renamed here (by __init__())
        self.properties['template_name'] = StringProperty(
            'template_name', self, icinga2_simple_mapping=False, object_mapping_type=self.type)
        self.properties['register'] = BooleanProperty(
            'register', self, default=True, icinga2_simple_mapping=False)
        for name, prop in self._properties.items():
            self.properties[name] = prop['type'](
                name, self,
                **dict((key, value) for key, value in prop.items() if key != 'type')
            )
        self.variables = {}
        self.unknown_properties = {}

    def set_property(self, name, **info):
        # Name property is renamed as template_name to be more explicite
        if name == 'name':
            name = 'template_name'
        elif name in self.properties_alias:
            name = self.properties_alias[name]

        if name in self.properties:
            self.properties[name].set_info(**info)
            return True

        # Detect custom object variable
        if name.startswith('_'):
            # Remove heading backspace in name
            self.add_variable(name[1:], icinga1_name=name, **info)
            return True
        return False

    def add_variable(self, name, icinga1_name=None, array=False, **info):
        var_type = IcingaObjectArrayVariable if array else IcingaObjectVariable
        self.variables[name] = var_type(
            icinga1_name if icinga1_name else name, self,
            icinga2_name=map_icinga1_custom_variable_name(name)
        )
        self.variables[name].set_info(**info)

    def __getattr__(self, name):
        if name == 'name':
            return self.get_name()
        if name == 'icinga2_name':
            return self.get_icinga2_name()
        if name in self.properties:
            return self.properties[name].get_value()
        raise IcingaObjectUnknownProperty(name)

    def get_name(self):
        return self.properties[self.name_property].get_value()

    def get_icinga2_name(self):
        return map_icinga1_object(self.type, self.name)

    def __repr__(self):
        return (
            f'<{type(self).__name__} {self.name} ({"template, " if not self.register else ""}'
            f'{self.filepath}:{self.start_line})>'
        )

    def format(self):
        lines = ["Object "]
        if self.template_name:
            lines[0] += "template "

        lines[0] += f'{self.type} {self.name}:'
        lines.append(
            f'  Defined in file {self.filepath} from line {self.start_line} to {self.end_line}')

        lines.append("  Properties:")
        for name, prop in self.properties.items():
            if prop.is_defined():
                lines.append(f'    - {name}: {prop}')
            elif self.templates:
                for template in self.templates:
                    if template.is_defined(name, use_template=True):
                        lines.append(
                            f'    - {name}: {template.getattr(name)}'
                            f' (from template {template.template_name})')
                        break
            else:
                lines.append(f'    - {name}: undefined')
        return "\n".join(lines)

    def set_template(self, templates):
        self.templates = []
        for template in self.use:
            if template not in templates:
                raise IcingaObjectUnknownTemplate(self)
            self.templates.append(templates[template])

    def is_defined(self, name, use_template=False):
        if name in self.properties and self.properties[name].is_defined():
            return True
        if use_template:
            for template in self.templates:
                if template.is_defined(name, use_template=True):
                    return True
        return False

    def is_excluded(self):
        return is_icinga1_excluded_object(self.type, self.name)

    def check_required(self):
        if not self.register:
            return True
        for name in self.properties:  # pylint: disable=consider-using-dict-items
            self.properties[name].check_required(self)
        return True


class IcingaHost(IcingaObject):

    type = 'host'
    name_property = 'host_name'

    _properties = {
        'host_name': dict(
            type=StringProperty, required=True, icinga2_simple_mapping=False,
            object_mapping_type='host'),
        'alias': dict(
            type=DisplayNameProperty, icinga2_name='display_name', icinga2_simple_mapping=False),
        'display_name': dict(type=DisplayNameProperty, icinga2_simple_mapping=False),
        'address': dict(type=AddressProperty, required=True),
        'parents': dict(
            type=CommaSeparatedListProperty, icinga2_simple_mapping=False,
            object_mapping_type='host'),
        'check_command': dict(type=CommandProperty, icinga2_command_type='CheckCommand'),
        'hostgroups': dict(
            type=CommaSeparatedListProperty, icinga2_name="groups",
            object_mapping_type='hostgroup'),
        'check_interval': dict(type=TimeUnitProperty),
        'retry_interval': dict(type=TimeUnitProperty),
        'check_period': dict(type=StringProperty, required=True, object_mapping_type='timeperiod'),
        'contacts': dict(
            type=CommaSeparatedListProperty, required=True,
            required_alt_properties=['contact_groups'], icinga2_simple_mapping=False),
        'contact_groups': dict(
            type=CommaSeparatedListProperty, required=True,
            required_alt_properties=['contacts'], icinga2_simple_mapping=False),
        'notification_interval': dict(
            type=TimeUnitProperty, required=True, icinga2_simple_mapping=False,
            icinga2_name='interval'),
        'notification_period': dict(
            type=StringProperty, required=True, icinga2_simple_mapping=False,
            icinga2_name='period', object_mapping_type='timeperiod'),
        'notifications_enabled': dict(
            type=BooleanProperty, required=True, icinga2_name='enable_notifications'),
        'notification_options': dict(
            type=HostNotificationOptionsProperty, icinga2_simple_mapping=False),
        'first_notification_delay': dict(
            type=TimeUnitProperty, icinga2_simple_mapping=False, icinga2_name='delay'),
        'event_handler_enabled': dict(type=BooleanProperty, icinga2_name="enable_event_handler"),
        'event_handler': dict(type=CommandProperty, icinga2_command_type='EventCommand'),
        'flap_detection_enabled': dict(
            type=BooleanProperty, default=True, icinga2_name="enable_flapping"),
        'failure_prediction_enabled': dict(type=BooleanProperty, icinga2_obsolete=True),
        'retain_status_information': dict(
            type=BooleanProperty, default=True, icinga2_obsolete=True),
        'retain_nonstatus_information': dict(
            type=BooleanProperty, default=True, icinga2_obsolete=True),
        'process_perf_data': dict(type=BooleanProperty, icinga2_name="enable_perfdata"),
        'max_check_attempts': dict(type=IntegerProperty),
        'notes': dict(type=StringProperty),
        'action_url': dict(type=StringProperty),
        'notes_url': dict(type=StringProperty),
        'icon_image': dict(type=StringProperty),
        'icon_image_alt': dict(type=StringProperty),
        'statusmap_image': dict(type=StringProperty, icinga2_obsolete=True),
        '2d_coords': dict(type=StringProperty, icinga2_obsolete=True),
        'vrml_image': dict(type=StringProperty, icinga2_obsolete=True),
    }


class IcingaHostGroup(IcingaObject):

    type = 'hostgroup'
    name_property = 'hostgroup_name'

    _properties = {
        'hostgroup_name': dict(
            type=StringProperty, required=True, icinga2_simple_mapping=False,
            object_mapping_type='hostgroup'),
        'alias': dict(type=DisplayNameProperty, icinga2_name="display_name"),
        'members': dict(type=HostMembersProperty),
        'hostgroup_members': dict(type=HostGroupMembersProperty, icinga2_name="groups"),
        'notes': dict(type=StringProperty, icinga2_unsupported=True),
    }


class IcingaService(IcingaObject):

    type = 'service'
    name_property = 'service_description'

    properties_alias = {
        'normal_check_interval': 'check_interval',
        'retry_check_interval': 'retry_interval',
        'hosts': 'host_name',
    }

    _properties = {
        'host_name': dict(
            type=HostMembersProperty, required=True, required_alt_properties=['hostgroup_name']),
        'hostgroup_name': dict(
            type=HostGroupMembersProperty, required=True, required_alt_properties=['host_name']),
        'service_description': dict(
            type=StringProperty, icinga2_simple_mapping=False, object_mapping_type='service'),
        'display_name': dict(type=DisplayNameProperty),
        'servicegroups': dict(
            type=CommaSeparatedListProperty, icinga2_name='groups',
            object_mapping_type='servicegroup'),
        'is_volatile': dict(type=BooleanProperty, default=False, icinga2_name='volatile'),
        'check_command': dict(
            type=CommandProperty, required=True, icinga2_command_type='CheckCommand'),
        'initial_state': dict(type=StringProperty, icinga2_obsolete=True),
        'max_check_attempts': dict(type=IntegerProperty),
        'check_interval': dict(type=TimeUnitProperty),
        'retry_interval': dict(type=TimeUnitProperty),
        'active_checks_enabled': dict(
            type=BooleanProperty, default=True, icinga2_name='enable_active_checks'),
        'passive_checks_enabled': dict(
            type=BooleanProperty, default=True, icinga2_name='enable_passive_checks'),
        'check_period': dict(type=StringProperty, required=True, object_mapping_type='timeperiod'),
        'obsess_over_service': dict(type=BooleanProperty, icinga2_obsolete=True),
        'check_freshness': dict(
            type=BooleanProperty, default=True, icinga2_obsolete=True),
        'freshness_threshold': dict(type=TimeUnitProperty, default=True, icinga2_obsolete=True),
        'event_handler': dict(
            type=CommandProperty, default=True, icinga2_command_type='EventCommand'),
        'event_handler_enabled': dict(type=BooleanProperty, icinga2_name="enable_event_handler"),
        'low_flap_threshold': dict(type=FloatProperty, icinga2_name='flapping_threshold_low'),
        'high_flap_threshold': dict(type=FloatProperty, icinga2_name='flapping_threshold_high'),
        'flap_detection_enabled': dict(
            type=BooleanProperty, default=True, icinga2_name='enable_flapping'),
        'flap_detection_options': dict(type=CommaSeparatedListProperty, icinga2_obsolete=True),
        'failure_prediction_enabled': dict(type=BooleanProperty, icinga2_obsolete=True),
        'process_perf_data': dict(
            type=BooleanProperty, default=True, icinga2_name='enable_perfdata'),
        'retain_status_information': dict(type=BooleanProperty, icinga2_obsolete=True),
        'retain_nonstatus_information': dict(type=BooleanProperty, icinga2_obsolete=True),
        'notification_interval': dict(
            type=TimeUnitProperty, icinga2_simple_mapping=False, icinga2_name='interval'),
        'first_notification_delay': dict(
            type=TimeUnitProperty, icinga2_simple_mapping=False, icinga2_name='delay'),
        'notification_period': dict(
            type=StringProperty, required=True, icinga2_simple_mapping=False,
            icinga2_name='period', object_mapping_type='timeperiod'),
        'notification_options': dict(
            type=ServiceNotificationOptionsProperty, icinga2_simple_mapping=False),
        'notifications_enabled': dict(
            type=BooleanProperty, required=True, icinga2_name='enable_notifications'),
        'contacts': dict(
            type=CommaSeparatedListProperty, required=True,
            required_alt_properties=['contact_groups'], icinga2_simple_mapping=False),
        'contact_groups': dict(
            type=CommaSeparatedListProperty, required=True,
            required_alt_properties=['contacts'], icinga2_simple_mapping=False),
        'stalking_options': dict(type=CommaSeparatedListProperty, icinga2_unsupported=True),
        'notes': dict(type=StringProperty),
        'notes_url': dict(type=StringProperty),
        'action_url': dict(type=StringProperty),
        'icon_image': dict(type=StringProperty),
        'icon_image_alt': dict(type=StringProperty),
    }


class IcingaServiceGroup(IcingaObject):

    type = 'servicegroup'
    name_property = 'servicegroup_name'

    _properties = {
        'servicegroup_name': dict(
            type=StringProperty, required=True, icinga2_simple_mapping=False,
            object_mapping_type='servicegroup'),
        'alias': dict(type=DisplayNameProperty, icinga2_name='display_name'),
        'members': dict(type=ObjectServiceGroupMembersProperty),
        'servicegroup_members': dict(type=ServiceGroupMembersProperty, icinga2_name='groups'),
        'notes': dict(type=StringProperty, icinga2_unsupported=True),
    }


class IcingaHostDependency(IcingaObject):

    type = 'hostdependency'

    _properties = {
        'host_name': dict(
            type=CommaSeparatedListProperty, required=True, icinga2_name='parent_host_name',
            required_alt_properties=['hostgroup_name'], icinga2_simple_mapping=False),
        'hostgroup_name': dict(
            type=CommaSeparatedListProperty, required=True, required_alt_properties=['host_name'],
            icinga2_unsupported=True, object_mapping_type='hostgroup'),
        'dependent_host_name': dict(
            type=HostMembersProperty, required=True,
            required_alt_properties=['dependent_hostgroup_name'], icinga2_name='child_host_name'),
        'dependent_hostgroup_name': dict(
            type=HostGroupMembersProperty, required=True,
            required_alt_properties=['dependent_host_name']),
        'inherits_parent': dict(type=InheritsParentProperty, default=False),
        'execution_failure_criteria': dict(type=HostFailureCriteriaProperty, icinga2_name='states'),
        'notification_failure_criteria': dict(
            type=NotificationFailureCriteriaProperty),
        'dependency_period': dict(
            type=StringProperty, icinga2_name='period', object_mapping_type='timeperiod'),
    }

    def get_name(self):
        childs = []
        if self.dependent_host_name:
            childs.extend([
                map_icinga1_object('host', host)
                for host in self.dependent_host_name
            ])
        if self.dependent_hostgroup_name:
            childs.extend([
                f'group-{map_icinga1_object("hostgroup", group)}'
                for group in self.dependent_hostgroup_name
            ])

        parents = []
        if self.host_name:
            parents.extend([
                map_icinga1_object('host', host)
                for host in self.host_name
            ])
        if self.hostgroup_name:
            parents.extend([
                f'group-{map_icinga1_object("hostgroup", group)}'
                for group in self.hostgroup_name
            ])

        return (
            f'{"+".join(childs) if childs else "?"}-depend-on-'
            f'{"+".join(parents) if parents else "?"}'
        ).replace(' ', '_').replace('!', 'not-')


class IcingaServiceDependency(IcingaObject):

    type = 'servicedependency'

    _properties = {
        'host_name': dict(
            type=CommaSeparatedListProperty, required=True, icinga2_name='parent_host_name',
            required_alt_properties=['hostgroup_name'], icinga2_simple_mapping=False),
        'hostgroup_name': dict(
            type=CommaSeparatedListProperty, required=True, required_alt_properties=['host_name'],
            icinga2_unsupported=True, object_mapping_type='hostgroup'),
        'service_description': dict(
            type=ServiceDescriptionProperty, required=True, icinga2_name='parent_service_name'),
        'dependent_host_name': dict(
            type=HostMembersProperty, required=True, icinga2_simple_mapping=False,
            required_alt_properties=['dependent_hostgroup_name'], icinga2_name='child_host_name'),
        'dependent_hostgroup_name': dict(
            type=HostGroupMembersProperty, required=True, icinga2_simple_mapping=False,
            required_alt_properties=['dependent_host_name']),
        'dependent_service_description': dict(
            type=ServiceDescriptionProperty, required=True, icinga2_simple_mapping=False,
            required_alt_properties=['dependent_servicegroup_name'],
            icinga2_name='child_service_name'),
        'dependent_servicegroup_name': dict(
            type=ServiceGroupMembersProperty, required=True,
            required_alt_properties=['dependent_service_description']),
        'inherits_parent': dict(type=InheritsParentProperty, default=False),
        'execution_failure_criteria': dict(
            type=ServiceFailureCriteriaProperty, icinga2_name='states'),
        'notification_failure_criteria': dict(
            type=NotificationFailureCriteriaProperty),
        'dependency_period': dict(
            type=StringProperty, icinga2_name='period', object_mapping_type='timeperiod'),
    }

    def get_name(self):
        child_hosts = []
        if self.dependent_host_name:
            child_hosts.extend([
                map_icinga1_object('host', host)
                for host in self.dependent_host_name
            ])
        if self.dependent_hostgroup_name:
            child_hosts.extend([
                f'group-{map_icinga1_object("hostgroup", group)}'
                for group in self.dependent_hostgroup_name
            ])
        child_hosts = '+'.join(child_hosts)

        child_services = []
        if self.dependent_service_description:
            child_services.append(
                map_icinga1_object('service', self.dependent_service_description)
            )
        if self.dependent_servicegroup_name:
            child_services.extend([
                f'group-{map_icinga1_object("servicegroup", group)}'
                for group in self.dependent_servicegroup_name
            ])
        child_services = '+'.join(child_services)

        if child_hosts and child_services:
            childs = f'{child_hosts}.{child_services}'
        elif child_hosts:
            childs = child_hosts
        elif child_services:
            childs = child_services
        else:
            childs = '?'

        parent_hosts = []
        if self.host_name:
            parent_hosts.extend([
                map_icinga1_object('host', host)
                for host in self.host_name
            ])
        if self.hostgroup_name:
            parent_hosts.extend([
                f'group-{map_icinga1_object("hostgroup", group)}'
                for group in self.hostgroup_name
            ])
        parent_hosts = '+'.join(parent_hosts)

        if parent_hosts and self.service_description:
            parents = f'{parent_hosts}.{map_icinga1_object("service", self.service_description)}'
        elif parent_hosts:
            parents = parent_hosts
        elif self.service_description:
            parents = map_icinga1_object("service", self.service_description)
        else:
            parents = '?'

        return f'{childs}-depend-on-{parents}'.replace(' ', '_').replace('!', 'not-')


class IcingaContact(IcingaObject):

    type = 'contact'
    name_property = 'contact_name'

    _properties = {
        'contact_name': dict(type=StringProperty, required=True, icinga2_simple_mapping=False),
        'alias': dict(type=DisplayNameProperty, icinga2_name='display_name'),
        'contactgroups': dict(type=CommaSeparatedListProperty, icinga2_name='groups'),
        'host_notifications_enabled': dict(
            type=BooleanProperty, default=True, icinga2_name='enable_notifications'),
        'service_notifications_enabled': dict(
            type=BooleanProperty, default=True, icinga2_simple_mapping=False),
        'host_notification_period': dict(
            type=StringProperty, required=True, icinga2_name='period',
            object_mapping_type='timeperiod'),
        'service_notification_period': dict(
            type=StringProperty, required=True, icinga2_simple_mapping=False,
            object_mapping_type='timeperiod'),
        'host_notification_options': dict(
            type=HostNotificationOptionsProperty, required=True, icinga2_simple_mapping=False),
        'service_notification_options': dict(
            type=ServiceNotificationOptionsProperty, required=True, icinga2_simple_mapping=False),
        'host_notification_commands': dict(
            type=CommaSeparatedListProperty, required=True, object_mapping_type='command',
            icinga2_unsupported=True),
        'service_notification_commands': dict(
            type=CommaSeparatedListProperty, required=True, object_mapping_type='command',
            icinga2_unsupported=True),
        'email': dict(type=StringProperty),
        'pager': dict(type=StringProperty),
    }


class IcingaContactGroup(IcingaObject):

    type = 'contactgroup'
    name_property = 'contactgroup_name'

    _properties = {
        'contactgroup_name': dict(type=StringProperty, required=True, icinga2_simple_mapping=False),
        'alias': dict(type=DisplayNameProperty, icinga2_name='display_name'),
        'members': dict(type=ContactMembersProperty),
        'contactgroup_members': dict(type=ContactGroupMembersProperty),
    }


class IcingaTimePeriod(IcingaObject):

    type = 'timeperiod'
    name_property = 'timeperiod_name'

    _properties = {
        'timeperiod_name': dict(
            type=StringProperty, required=True, icinga2_simple_mapping=False,
            object_mapping_type='timeperiod'),
        'alias': dict(type=DisplayNameProperty, icinga2_name='display_name'),
        'sunday': dict(type=TimePeriodWeekDayProperty, icinga2_simple_mapping=False),
        'monday': dict(type=TimePeriodWeekDayProperty, icinga2_simple_mapping=False),
        'tuesday': dict(type=TimePeriodWeekDayProperty, icinga2_simple_mapping=False),
        'wednesday': dict(type=TimePeriodWeekDayProperty, icinga2_simple_mapping=False),
        'thursday': dict(type=TimePeriodWeekDayProperty, icinga2_simple_mapping=False),
        'friday': dict(type=TimePeriodWeekDayProperty, icinga2_simple_mapping=False),
        'saturday': dict(type=TimePeriodWeekDayProperty, icinga2_simple_mapping=False),
        'exclude': dict(
            type=CommaSeparatedListProperty, icinga2_simple_mapping=True, icinga2_name='excludes'),
    }

    # pylint: disable=line-too-long
    _custom_week_day_property_name = re.compile(
        r'^(((sunday|monday|tuesday|wednesday|thursday|friday|saturday)(\s+[0-9]+(\s+(january|february|march|april|may|june|july|august|september|october|november|december))?)?)|([0-9]{4}\-[0-9]{1,2}\-[0-9]{1,2}(\s-\s[0-9]{4}\-[0-9]{1,2}\-[0-9]{1,2})?)|((day|january|february|march|april|may|june|july|august|september|october|november|december)\s+{0-9}+(\s+-\s+-?{0-9}+)?)|([0-9]+\s+(january|february|march|april|may|june|july|august|september|october|november|december))|((january|february|march|april|may|june|july|august|september|october|november|december)\s+[0-9]+))$')

    def set_property(self, name, **info):
        if super().set_property(name, **info):
            return True
        if self._custom_week_day_property_name.match(name):
            self.properties[name] = TimePeriodWeekDayProperty(
                name, self, icinga2_simple_mapping=False
            )
            self.properties[name].set_info(**info)
            return True
        return False


class IcingaCommand(IcingaObject):

    type = 'command'
    name_property = 'command_name'

    _properties = {
        'command_name': dict(
            type=StringProperty, required=True, icinga2_simple_mapping=False,
            object_mapping_type='command'),
        'command_line': dict(type=CommandLineProperty, required=True),
    }
