""" Icinga objects properties """

import logging
import re
import shlex

from .exceptions import IcingaObjectMissingProperty
from .icinga2 import (
    escape_string, map_icinga1_object, map_icinga1_object_display_name, map_icinga1_macro,
    map_icinga1_custom_variable_name, is_ignored_icinga1_unsupported_property)


log = logging.getLogger(__name__)


class IcingaObjectProperty:
    """ Icinga object property common class """

    name = None
    icinga2_name = None
    icinga2_unsupported = False
    icinga2_obsolete = False
    icinga2_simple_mapping = True
    required = False
    default = None
    required_alt_properties = []
    value = None
    comment = None
    pre_comment_lines = []
    always_exported = False

    def __init__(self, name, obj, default=None, **kwargs):
        self.name = name
        self.obj = obj
        if default is not None:
            self.default = default
        self.set_info(**kwargs)

    def set_value(self, value):
        self.value = value

    def get_value(self):
        if self.is_defined():
            return self.value
        return self.default

    def get_display_value(self):
        return self.get_value()

    _quotted_need = re.compile('[^a-zA-Z0-9_]')

    def get_icinga2_name(self, quotted=False):
        name = self.icinga2_name if self.icinga2_name else self.name
        if name and quotted and self._quotted_need.search(name):
            name = f'"{name}"'
        return name

    def get_icinga2_value(self):
        return self.get_value()

    # pylint: disable=unused-argument
    def get_icinga2_directive(self, prefix=None, object_type=None, raw_values=False):
        prefix = prefix if prefix else ""
        lines = [
            f'{prefix}# {line}'
            for line in self.pre_comment_lines
        ]
        lines.append(
            f'{prefix}{self.get_icinga2_name(quotted=True)} = {self.get_icinga2_value()}'
            f'{(" # " + self.comment) if self.comment else ""}'
        )
        return "\n".join(lines)

    def is_obsolete_in_icinga2(self):
        return self.icinga2_obsolete

    def is_supported_by_icinga2(self):
        return not self.icinga2_unsupported and not self.icinga2_obsolete

    def is_ignored_icinga1_unsupported_property(self):
        return is_ignored_icinga1_unsupported_property(self.name)

    def is_defined(self):
        return self.value is not None

    def set_comment(self, comment):
        self.comment = comment

    def set_pre_comment_lines(self, pre_comment_lines):
        self.pre_comment_lines = pre_comment_lines

    def set_info(self, **info):
        for arg, value in info.items():
            if arg == 'value':
                self.set_value(value)
            elif arg == 'comment':
                self.set_comment(value)
            elif arg == 'pre_comment_lines':
                self.set_pre_comment_lines(value)
            else:
                assert hasattr(self, arg), (
                    f'{self}: {type(self).__name__} have no attribute "{arg}" (value={value})'
                )
                setattr(self, arg, value)

    def __repr__(self):
        return (
            f'<{type(self).__name__} {self.name} = {self.get_display_value()}'
            f'{("# " + self.comment) if self.comment else ""}>'
        )

    def check_required(self, obj):
        if obj.is_defined(self.name, use_template=True) or not self.required:
            return True
        if self.required_alt_properties:
            for alt_p in self.required_alt_properties:
                if obj.is_defined(alt_p, use_template=True):
                    return True
        raise IcingaObjectMissingProperty(
            self.name, alt_properties=self.required_alt_properties)

    def __copy__(self):
        return type(self)(
            self.name, self, default=self.default,
            required=self.required, required_alt_properties=self.required_alt_properties
        )


class StringProperty(IcingaObjectProperty):

    object_mapping_type = None

    def get_icinga2_value(self):
        value = self.get_value()
        if self.object_mapping_type:
            value = map_icinga1_object(self.object_mapping_type, value)
        return f'"{escape_string(map_icinga1_macro(value))}"'


class DisplayNameProperty(StringProperty):

    def get_icinga2_value(self):
        value = map_icinga1_object_display_name(
            self.object_mapping_type if self.object_mapping_type else self.obj.type,
            self.get_value()
        )
        return f'"{escape_string(map_icinga1_macro(value))}"'


class AddressProperty(StringProperty):

    # pylint: disable=line-too-long
    _is_ipv6_address = re.compile('^(?:(?:[0-9A-Fa-f]{1,4}:){6}(?:[0-9A-Fa-f]{1,4}:[0-9A-Fa-f]{1,4}|(?:(?:[0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])\\.){3}(?:[0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5]))|::(?:[0-9A-Fa-f]{1,4}:){5}(?:[0-9A-Fa-f]{1,4}:[0-9A-Fa-f]{1,4}|(?:(?:[0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])\\.){3}(?:[0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5]))|(?:[0-9A-Fa-f]{1,4})?::(?:[0-9A-Fa-f]{1,4}:){4}(?:[0-9A-Fa-f]{1,4}:[0-9A-Fa-f]{1,4}|(?:(?:[0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])\\.){3}(?:[0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5]))|(?:[0-9A-Fa-f]{1,4}:[0-9A-Fa-f]{1,4})?::(?:[0-9A-Fa-f]{1,4}:){3}(?:[0-9A-Fa-f]{1,4}:[0-9A-Fa-f]{1,4}|(?:(?:[0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])\\.){3}(?:[0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5]))|(?:(?:[0-9A-Fa-f]{1,4}:){,2}[0-9A-Fa-f]{1,4})?::(?:[0-9A-Fa-f]{1,4}:){2}(?:[0-9A-Fa-f]{1,4}:[0-9A-Fa-f]{1,4}|(?:(?:[0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])\\.){3}(?:[0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5]))|(?:(?:[0-9A-Fa-f]{1,4}:){,3}[0-9A-Fa-f]{1,4})?::[0-9A-Fa-f]{1,4}:(?:[0-9A-Fa-f]{1,4}:[0-9A-Fa-f]{1,4}|(?:(?:[0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])\\.){3}(?:[0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5]))|(?:(?:[0-9A-Fa-f]{1,4}:){,4}[0-9A-Fa-f]{1,4})?::(?:[0-9A-Fa-f]{1,4}:[0-9A-Fa-f]{1,4}|(?:(?:[0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])\\.){3}(?:[0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5]))|(?:(?:[0-9A-Fa-f]{1,4}:){,5}[0-9A-Fa-f]{1,4})?::[0-9A-Fa-f]{1,4}|(?:(?:[0-9A-Fa-f]{1,4}:){,6}[0-9A-Fa-f]{1,4})?::)$')

    def get_icinga2_name(self, quotted=False):
        if self._is_ipv6_address.match(self.get_value()):
            return "address6"
        return "address"


class CommandProperty(IcingaObjectProperty):

    command = None
    args = []
    icinga2_command_type = None

    _detect_extra_command_args = re.compile(
        r'[^a-zA-Z0-9](\-[a-zA-Z0-9\-\_]+)')

    @classmethod
    def _parse_command_args(cls, command):
        return [part.strip() for part in command.split('!')]

    def set_value(self, value):
        self.value = value
        self.args = self._parse_command_args(value)
        self.command = self.args.pop(0)

    def get_icinga2_directive(self, prefix=None, object_type=None, raw_values=False):
        if raw_values:
            return super().get_icinga2_directive(
                prefix=prefix,
                object_type=object_type
            )
        assert self.icinga2_command_type, f'{self}: You must define icinga2_command_type attribute'
        prefix = prefix if prefix else ""
        lines = [
            f'{prefix}{self.get_icinga2_name(quotted=True)} = '
            f'"{escape_string(map_icinga1_object("command", self.command))}"'
        ]
        for idx, arg in enumerate(self.args):
            arg_value = map_icinga1_macro(arg)
            # Detect extra commands args in arg value (ex: -w or --warning)
            extra_command_args = self._detect_extra_command_args.findall(arg_value)
            if extra_command_args:
                log.warning(
                    'Potential extra command arguments detected in %s ("%s"): '
                    'you potentialy need to convert it manually as extra command variable(s)',
                    self, '", "'.join(extra_command_args)
                )
            var_name = map_icinga1_custom_variable_name(
                f'{self.icinga2_command_type.upper()}_ARG{idx + 1}'
            )
            lines.append(f'{prefix}vars.{var_name} = "{escape_string(arg_value)}"')
        return "\n".join(lines)


class CommandLineProperty(IcingaObjectProperty):

    command = None
    args = []

    def set_value(self, value):
        self.value = value
        try:
            args = shlex.split(value)
        except ValueError:
            log.exception(
                '%s: Exception occured spliting args from value "%s"',
                self, value)
            raise
        log.debug("%s: Split command line '%s' : %s", self, value, args)
        self.command = [args.pop(0)]
        while True:
            if not args or args[0].startswith('-'):
                break
            self.command.append(args.pop(0))

        # Parse arguments
        self.args = []
        current_arg = {}
        for arg in args:
            if arg.startswith('-'):
                if current_arg:
                    self.args.append(current_arg)
                current_arg = {
                    'key': arg,
                    'value': None
                }
            elif current_arg:
                current_arg["value"] = arg
                self.args.append(current_arg)
                current_arg = {}
            else:
                self.args.append({
                    'key': None,
                    'value': arg
                })
        if current_arg:
            self.args.append(current_arg)
        log.debug('%s.set_value(%s): args=%s', self, value, self.args)

    _is_command_arg = re.compile(r'^\$ARG[0-9]+\$$')

    @classmethod
    def is_required_arg(cls, key, value):
        if not key and cls._is_command_arg.match(value):
            return False
        return True

    def get_icinga2_directive(self, prefix=None, object_type=None, raw_values=False):
        if raw_values:
            return super().get_icinga2_directive(
                prefix=prefix,
                object_type=object_type
            )
        if prefix is None:
            prefix = ""
        lines = []
        if self.pre_comment_lines:
            lines += [
                f'{prefix}# {line}'
                for line in self.pre_comment_lines
            ]
        if self.comment:
            lines.append(f'{prefix}# {self.comment}')
        lines.append(
            f'{prefix}# Icinga1 {self.name}: {self.value}'
        )

        lines.append(
            '%scommand = [ "%s" ]' % (  # pylint: disable=consider-using-f-string
                prefix,
                '", "'.join(
                    [
                        map_icinga1_macro(
                            arg,
                            map_command_args=True,
                            map_command_type=object_type
                        )
                        for arg in self.command
                    ]
                )
            )
        )

        if self.args:
            args_keys = []
            duplicated_keys = {}
            lines.append(prefix + "arguments = {")
            for idx, arg in enumerate(self.args):
                ckey = arg.get('key')
                cvalue = arg.get('value')
                if not ckey and not cvalue:
                    continue
                required = self.is_required_arg(ckey, cvalue)
                skip_key = False
                if not ckey:
                    ckey = f'--{cvalue.replace("$", "").lower()}'
                    skip_key = True
                if not cvalue:
                    cvalue = ckey
                    skip_key = True
                cvalue = map_icinga1_macro(
                    cvalue,
                    map_command_args=True,
                    map_command_type=object_type
                )

                # Handle duplicated key
                if ckey in args_keys:
                    if ckey not in duplicated_keys:
                        log.debug('%s: duplicated %s argument detected', self, ckey)
                        duplicated_keys[ckey] = 1
                    duplicated_keys[ckey] += 1
                    key_name = f'{ckey}{duplicated_keys[ckey]}'
                else:
                    args_keys.append(ckey)
                    key_name = ckey

                # pylint: disable=consider-using-f-string
                lines.append('%s  "%s" = {' % (prefix, escape_string(key_name)))
                lines.append(f'{prefix}    value = "{escape_string(cvalue)}"')
                if skip_key:
                    lines.append(f'{prefix}    skip_key = true')
                elif key_name != ckey:
                    lines.append(f'{prefix}    key = "{escape_string(ckey)}"')
                if required:
                    lines.append(f'{prefix}    required = true')
                lines.append(f'{prefix}    order = {idx}')
                lines.append(prefix + '  }')
            lines.append(prefix + "}")
        return "\n".join(lines)


class IntegerProperty(IcingaObjectProperty):
    def set_value(self, value):
        self.value = int(value)


class FloatProperty(IcingaObjectProperty):
    def set_value(self, value):
        self.value = float(value)


class TimeUnitProperty(IntegerProperty):

    interval_length = None

    def set_value(self, value):
        # pylint: disable=import-outside-toplevel
        from .icinga2 import icinga1_interval_length
        interval_length = (
            self.interval_length if self.interval_length
            else icinga1_interval_length)
        self.value = int(value) * interval_length  # Store in seconds

    def get_display_value(self):
        return self.get_icinga2_value()

    def get_icinga2_value(self):
        value = self.get_value()
        log.debug(
            "%s.get_icinga2_value(%s): %s",
            type(self).__name__, value, type(value))
        if not value:
            return value
        result = []
        if value > 86400:
            result.append(f'{value // 86400}d')
            value = value % 86400
        if value > 3600:
            result.append(f'{value // 3600}h')
            value = value % 3600
        if value > 60:
            result.append(f'{value // 60}m')
            value = value % 60
        if value:
            result.append(f'{value}s')
        return ' + '.join(result)


class BooleanProperty(IcingaObjectProperty):
    def set_value(self, value):
        self.value = bool(int(value))

    def get_icinga2_value(self):
        if self.get_value():
            return "true"
        return "false"


class CommaSeparatedListProperty(IcingaObjectProperty):

    default = []
    can_contain_macro = False
    object_mapping_type = None
    _auto_object_mapping = True

    def set_value(self, value):
        self.value = [v.strip() for v in value.split(',')]

    def get_icinga2_value(self):
        values = self.get_value()
        if self.object_mapping_type and self._auto_object_mapping:
            values = [
                map_icinga1_object(self.object_mapping_type, value)
                for value in values
            ]
        if self.can_contain_macro:
            values = [map_icinga1_macro(value) for value in values]
        return '["%s"]' % '", "'.join(  # pylint: disable=consider-using-f-string
            [
                escape_string(value)
                for value in values
            ]
        )


class UseProperty(CommaSeparatedListProperty):

    def get_icinga2_directive(self, prefix=None, object_type=None, raw_values=False):
        if raw_values:
            return super().get_icinga2_directive(
                prefix=prefix,
                object_type=object_type
            )
        if not prefix:
            prefix = ""
        lines = []
        if self.pre_comment_lines:
            lines += [
                f'{prefix}# {line}'
                for line in self.pre_comment_lines
            ]
        if self.comment:
            lines.append(f'{prefix}# {self.comment}')
        for tpl in self.get_value():
            lines.append(f'{prefix}import "{escape_string(map_icinga1_object(object_type, tpl))}"')
        return "\n".join(lines)


class MembersProperty(CommaSeparatedListProperty):

    icinga2_member_obj_type = None
    _auto_object_mapping = False

    def get_icinga2_directive(self, prefix=None, object_type=None, raw_values=False,
                              additional_clause=None):
        if raw_values:
            return super().get_icinga2_directive(
                prefix=prefix,
                object_type=object_type
            )
        assert self.icinga2_member_obj_type is not None, \
            "You must set icinga2_member_obj_type property"

        included = []
        include_star = False
        excluded = []
        exclude_star = False
        for member in self.get_value():
            if member.startswith('!'):
                member = member[1:]
                if member == '*':
                    exclude_star = True
                else:
                    excluded.append(map_icinga1_object(self.icinga2_member_obj_type, member))
            else:
                if member == '*':
                    include_star = True
                else:
                    included.append(map_icinga1_object(self.icinga2_member_obj_type, member))

        lines = []
        if not prefix:
            prefix = ""
        if self.pre_comment_lines:
            lines += [
                f'{prefix}# {line}'

                for line in self.pre_comment_lines
            ]
        if include_star:
            lines.append(
                f'{prefix}assign where match("*", {self.icinga2_member_obj_type}.name)'
                f'{(" && " + additional_clause) if additional_clause else ""}'
            )
        if included:
            lines.append(
                f'{prefix}assign where {self.icinga2_member_obj_type}.name in '
                + '["%s"]' % '", "'.join(included)  # pylint: disable=consider-using-f-string
                + f'{(" && " + additional_clause) if additional_clause else ""}'
            )
        if exclude_star:
            lines.append(
                f'{prefix}ignore where match("*", {self.icinga2_member_obj_type}.name)'
                f'{(" && " + additional_clause) if additional_clause else ""}'
            )
        if excluded:
            lines.append(
                f'{prefix}ignore where {self.icinga2_member_obj_type}.name in '
                + '["%s"]' % '", "'.join(excluded)  # pylint: disable=consider-using-f-string
                + f'{(" && " + additional_clause) if additional_clause else ""}'
            )

        return "\n".join(lines)


class HostMembersProperty(MembersProperty):

    icinga2_member_obj_type = 'host'


class ContactMembersProperty(MembersProperty):

    icinga2_member_obj_type = 'user'


class MemberProperty(StringProperty):

    icinga2_member_obj_type = None
    _auto_object_mapping = False

    def get_icinga2_directive(self, prefix=None, object_type=None, raw_values=False):
        if raw_values:
            return super().get_icinga2_directive(
                prefix=prefix,
                object_type=object_type
            )
        assert self.icinga2_member_obj_type is not None, \
            "You must set icinga2_member_obj_type property"

        included = []
        include_star = False
        excluded = []
        exclude_star = False
        member = self.get_value()
        if member.startswith('!'):
            member = member[1:]
            excluded.append(map_icinga1_object(self.icinga2_member_obj_type, member))
        else:
            included.append(map_icinga1_object(self.icinga2_member_obj_type, member))

        lines = []
        if not prefix:
            prefix = ""
        if self.pre_comment_lines:
            lines += [
                f'{prefix}# {line}'
                for line in self.pre_comment_lines
            ]
        if include_star:
            lines.append(
                f'{prefix}assign where match("*", {self.icinga2_member_obj_type}.name)'
            )
        if included:
            lines.append(
                f'{prefix}assign where {self.icinga2_member_obj_type}.name in '
                '["%s"]' % '", "'.join(included)  # pylint: disable=consider-using-f-string
            )
        if exclude_star:
            lines.append(
                f'{prefix}ignore where match("*", {self.icinga2_member_obj_type}.name)'
            )
        if excluded:
            lines.append(
                f'{prefix}ignore where {self.icinga2_member_obj_type}.name in '
                '["%s"]' % '", "'.join(excluded)  # pylint: disable=consider-using-f-string
            )

        return "\n".join(lines)


class ObjectServiceGroupMembersProperty(CommaSeparatedListProperty):

    _auto_object_mapping = False

    def set_value(self, value):
        super().set_value(value)
        if not self.value:
            return True

        values = []
        while len(self.value) >= 2:
            host_name = self.value.pop(0)
            service_description = self.value.pop(0)
            values.append((host_name, service_description))
        if self.value:
            log.warning(
                '%s: Invalid number of comma separated parameter values '
                '(must be divisible by two): Ignore last value (%s).',
                self,
                self.value
            )
        self.value = values

    def get_icinga2_directive(self, prefix=None, object_type=None, raw_values=False):
        if raw_values:
            return super().get_icinga2_directive(
                prefix=prefix,
                object_type=object_type
            )
        lines = []
        if not prefix:
            prefix = ""
        if self.pre_comment_lines:
            lines += [
                f'{prefix}# {line}'
                for line in self.pre_comment_lines
            ]
        for host_name, service_description in self.get_value():
            service_description = map_icinga1_object('service', service_description)
            if host_name == '*':
                lines.append(
                    f'{prefix}assign where service.name == "{service_description}"'
                )
            else:
                host_name = map_icinga1_object('host', host_name)
                lines.append(
                    f'{prefix}assign where host.name == "{host_name}" '
                    f'&& service.name == "{service_description}"'
                )

        return "\n".join(lines)


class GroupMembersProperty(CommaSeparatedListProperty):

    icinga2_member_obj_type = None
    icinga2_root_member_obj_type = None
    _auto_object_mapping = False

    def get_icinga2_directive(self, prefix=None, object_type=None, raw_values=False,
                              additional_clause=None):
        if raw_values:
            return super().get_icinga2_directive(
                prefix=prefix,
                object_type=object_type
            )
        assert self.icinga2_member_obj_type is not None, \
            "You must set icinga2_member_obj_type property"
        assert self.icinga2_root_member_obj_type is not None, \
            "You must set icinga2_root_member_obj_type property"

        included = []
        include_star = False
        excluded = []
        exclude_star = False
        for member in self.get_value():
            if member.startswith('!'):
                member = member[1:]
                if member == '*':
                    exclude_star = True
                else:
                    excluded.append(map_icinga1_object(self.icinga2_member_obj_type, member))
            else:
                if member == '*':
                    include_star = True
                else:
                    included.append(map_icinga1_object(self.icinga2_member_obj_type, member))

        # Note: always handle members with assign & ignore rules
        # to permit usage of apply objects.
        lines = []
        if not prefix:
            prefix = ""
        if self.pre_comment_lines:
            lines += [
                f'{prefix}# {line}'
                for line in self.pre_comment_lines
            ]
        if include_star:
            lines.append(
                f'{prefix}assign where match("*", {self.icinga2_root_member_obj_type}.name)'
                f'{(" && " + additional_clause) if additional_clause else ""}'
            )

        for member in included:
            lines.append(
                f'{prefix}assign where "{member}" in {self.icinga2_root_member_obj_type}.groups'
                f'{(" && " + additional_clause) if additional_clause else ""}'
            )
        if exclude_star:
            lines.append(
                f'{prefix}ignore where match("*", {self.icinga2_root_member_obj_type}.name)'
                f'{(" && " + additional_clause) if additional_clause else ""}'
            )
        for member in excluded:
            lines.append(
                f'{prefix}ignore where "{member}" in {self.icinga2_root_member_obj_type}.groups'
                f'{(" && " + additional_clause) if additional_clause else ""}'
            )
        return "\n".join(lines)


class HostGroupMembersProperty(GroupMembersProperty):

    icinga2_member_obj_type = 'hostgroup'
    icinga2_root_member_obj_type = 'host'


class ServiceGroupMembersProperty(GroupMembersProperty):

    icinga2_member_obj_type = 'servicegroup'
    icinga2_root_member_obj_type = 'service'


class ContactGroupMembersProperty(GroupMembersProperty):

    icinga2_member_obj_type = 'usergroup'
    icinga2_root_member_obj_type = 'user'


class ServiceDescriptionProperty(StringProperty):

    object_mapping_type = 'service'

    def get_icinga2_directive(self, prefix=None, object_type=None, raw_values=False):
        if raw_values or self.get_value() != '*':
            return super().get_icinga2_directive(
                prefix=prefix, object_type=object_type, raw_values=raw_values
            )

        return ""


class HostNotificationOptionsProperty(CommaSeparatedListProperty):

    _auto_object_mapping = False

    available_options = {
        'd': 'DOWN',
        'u': 'UNREACHABLE',
        'r': 'RECOVERY',
        'f': 'FLAPPING',
        's': 'SCHEDULED_DOWNTIME',
        'n': 'NONE',
    }

    icinga2_states_mapping = {
        'd': ['Down'],
        'r': ['Up'],
        'f': ['Up', 'Down'],
    }

    icinga2_types_mapping = {
        'd': ['Problem'],
        'u': ['Problem'],
        'r': ['Recovery'],
        'f': ['FlappingStart', 'FlappingEnd'],
        's': ['DowntimeStart', 'DowntimeEnd', 'DowntimeRemoved'],
    }

    def set_value(self, value):
        self.value = []
        for option in [v.strip().lower() for v in value.split(',')]:
            if option not in self.available_options:
                log.warning('%s: Invalid option "%s"', self, option)
            self.value.append(option)

    def get_icinga2_states(self):
        values = self.get_value()
        # Handle special case 'n'
        if 'n' in values:
            return []
        states = []
        for value, value_states in self.icinga2_states_mapping.items():
            if value not in values:
                continue
            for state in value_states:
                if state not in states:
                    states.append(state)
        return states

    def get_icinga2_types(self):
        values = self.get_value()
        # Handle special case 'n'
        if 'n' in values:
            return []
        # Always add unmappable types
        types = ['Acknowledgement', 'Custom']
        for value, value_types in self.icinga2_types_mapping.items():
            if value not in values:
                continue
            for t in value_types:
                if t not in types:
                    types.append(t)
        return types

    def get_icinga2_directive(self, prefix=None, object_type=None, raw_values=False):
        if raw_values:
            return super().get_icinga2_directive(
                prefix=prefix,
                object_type=object_type
            )
        if not prefix:
            prefix = ""
        return "\n".join([
            super().get_icinga2_directive(
                prefix=prefix + '# ',
                object_type=object_type
            ),
            f'{prefix}states = [{", ".join(self.get_icinga2_states())}]',
            f'{prefix}types = [{", ".join(self.get_icinga2_types())}]'
        ])


class ServiceNotificationOptionsProperty(HostNotificationOptionsProperty):

    available_options = {
        'w': 'WARNING',
        'u': 'UNKNOWN',
        'c': 'CRITICAL',
        'r': 'RECOVERY',
        'f': 'FLAPPING',
        's': 'SCHEDULED_DOWNTIME',
        'n': 'NONE',
    }

    icinga2_states_mapping = {
        'w': ['Warning'],
        'u': ['Unknown'],
        'c': ['Critical'],
        'r': ['OK'],
        'f': ['Warning', 'Critical', 'Unknown', 'OK'],
    }

    icinga2_types_mapping = {
        'w': ['Problem'],
        'u': ['Problem'],
        'c': ['Problem'],
        'r': ['Recovery'],
        'f': ['FlappingStart', 'FlappingEnd'],
        's': ['DowntimeStart', 'DowntimeEnd', 'DowntimeRemoved'],
    }


class HostFailureCriteriaProperty(HostNotificationOptionsProperty):

    available_options = {
        'o': 'UP',
        'd': 'DOWN',
        'u': 'UNREACHABLE',
        'p': 'PENDING',
        'n': 'NONE',
    }

    icinga2_state_mapping = {
        'o': 'UP',
        'd': 'DOWN',
        'u': 'DOWN',
        'p': 'DOWN',
    }

    def get_icinga2_value(self):
        # execution_failure_criteria -> states conversion
        #   Icinga1 execution_failure_criteria = list of failed states
        #   Icinga2 states = list of OK states
        # => Inverse states list and keep only mappable states
        states = []
        values = self.get_value()
        # Particular case 'n'
        if 'n' in values:
            log.debug(
                '%s.get_icinga2_value(%s): Particular case: "n" => '
                'execution dependency will never fail',
                self,
                values
            )
            states.extend(self.icinga2_state_mapping.values())
        else:
            # Otherwise
            for opt in self.available_options:
                # Ignore unmappable options
                if opt not in self.icinga2_state_mapping:
                    continue
                # Include only state not in values
                if opt in values:
                    continue
                states.append(self.icinga2_state_mapping[opt])
        return f'[{", ".join(states)}]'

    def get_icinga2_directive(self, prefix=None, object_type=None, raw_values=False):
        lines = [
            super().get_icinga2_directive(
                prefix=prefix,
                object_type=object_type,
                raw_values=True
            )
        ]
        if not prefix or '#' not in prefix:
            lines.append(
                f'{prefix if prefix else ""}disable_checks = true'
            )
        return "\n".join(lines)


class ServiceFailureCriteriaProperty(HostFailureCriteriaProperty):

    available_options = {
        'o': 'OK',
        'w': 'WARNING',
        'u': 'UNKNOWN',
        'c': 'CRITICAL',
        'p': 'PENDING',
        'n': 'NONE',
    }

    icinga2_state_mapping = {
        'o': 'OK',
        'w': 'Warning',
        'u': 'Unknown',
    }


class NotificationFailureCriteriaProperty(CommaSeparatedListProperty):

    def get_icinga2_value(self):
        return None

    def get_icinga2_directive(self, prefix=None, object_type=None, raw_values=False):
        lines = []
        values = self.get_value()
        # Particular case 'n' : never disable notification
        if 'n' not in values:
            if not self.is_ignored_icinga1_unsupported_property():
                log.warning(
                    '%s: Unsupported Icinga1 %s property : Icinga2 do not permit to only disable '
                    'notification in certain states. However, the parameter states permit to set '
                    'when the dependency should be OK.', self.obj, self.name)
            lines += [
                f'{prefix if prefix else ""}# Unsupported Icinga1 {self.name} property:',
                f'{prefix if prefix else ""}# Icinga2 do not permit to only disable notification '
                'in certain states. However,',
                f'{prefix if prefix else ""}# the parameter states permit to'
                ' set when the dependency should be OK.',
            ]
        lines.append(
            f'{prefix if prefix else ""}disable_notifications = '
            f'{"false" if "n" in values else "true"}'
        )
        return "\n".join(lines)


class TimePeriodWeekDayProperty(IcingaObjectProperty):
    default = []

    # pylint: disable=line-too-long
    _parse_hour_period = re.compile(
        '^(?P<start_hour>[0-2]?[0-9]):(?P<start_min>[0-6]?[0-9])-(?P<end_hour>[0-2]?[0-9]):(?P<end_min>[0-6]?[0-9])$')

    def set_value(self, value):
        self.value = []
        for period in [v.strip() for v in value.split(',')]:
            m = self._parse_hour_period.match(period)
            if not m:
                log.warning('Invalid hour period "%s"', period)
                continue

            start_hour = int(m.group('start_hour'))
            if start_hour < 0 or start_hour > 23:
                log.warning('Invalid start hour in period "%s"', period)
                continue
            start_min = int(m.group('start_min'))
            if start_min < 0 or start_min > 59:
                log.warning('Invalid start min in period "%s"', period)
                continue
            end_hour = int(m.group('end_hour'))
            if end_hour < 0 or end_hour > 24:
                log.warning('Invalid end hour in period "%s"', period)
                continue
            end_min = int(m.group('end_min'))
            if end_min < 0 or end_min > 59:
                log.warning('Invalid end min in period "%s"', period)
                continue

            start = start_hour * 100 + start_min
            end = end_hour * 100 + end_min
            if start >= end and not (start == 0 and end == 0):
                log.warning('Invalid start/end hours in period "%s"', period)
                continue

            self.value.append(
                dict(
                    start_hour=start_hour,
                    start_min=start_min,
                    end_hour=end_hour,
                    end_min=end_min,
                )
            )

    def get_icinga2_name(self, quotted=False):
        name = super().get_icinga2_name(quotted=False)

        # Fix YYYY-M-D to YYYY-MM-DD
        for date in re.finditer(r'([0-9]{4})\-([0-9]{1,2})\-([0-9]{1,2})', name):
            name = name.replace(
                date.group(0),
                "%d-%.02d-%.02d" % (  # pylint: disable=consider-using-f-string
                    int(date.group(1)),
                    int(date.group(2)),
                    int(date.group(3)),
                )
            )

        if name and quotted and self._quotted_need.search(name):
            name = f'"{name}"'
        return name

    def get_icinga2_value(self):
        # pylint: disable=consider-using-f-string
        return '"%s"' % ','.join([
            "{start_hour}:{start_min}-{end_hour}:{end_min}".format(
                start_hour="%.02d" % p['start_hour'],
                start_min="%.02d" % p['start_min'],
                end_hour=(
                    "%.02d" % 24 if p['end_hour'] == 0 and p['end_min'] == 0 else p['end_hour']),
                end_min="%.02d" % p['end_min'],
            )
            for p in self.get_value()
        ])


class InheritsParentProperty(BooleanProperty):

    always_exported = True

    def get_icinga2_directive(self, prefix=None, object_type=None, raw_values=False):
        if raw_values:
            return super().get_icinga2_directive(
                prefix=prefix,
                object_type=object_type
            )
        # inherits_parent property is always True in Icinga2
        if self.get_value():
            return ''

        if not self.is_ignored_icinga1_unsupported_property():
            log.warning(
                '%s: Unsupported Icinga1 inherits_parent = 0 detected: Dependencies are always '
                'inherits from parent in Icinga2.', self.obj)
        prefix = prefix if prefix else ""
        return "\n".join([
            f'{prefix}# Icinga1 unsupported property: inherits_parent = 0',
            f'{prefix}# Dependencies are always inherits from parent in Icinga2.'
        ])


class IcingaObjectVariable(StringProperty):

    # pylint: disable=unused-argument
    def get_icinga2_directive(self, prefix=None, object_type=None, raw_values=False):
        prefix = prefix if prefix else ""
        lines = [
            f'{prefix}# {line}'
            for line in self.pre_comment_lines
        ]
        lines.append(
            f'{prefix}vars.{self.get_icinga2_name(quotted=True)} = {self.get_icinga2_value()}'
            f'{(" # " + self.comment) if self.comment else ""}'
        )
        return "\n".join(lines)


class IcingaObjectArrayVariable(CommaSeparatedListProperty, IcingaObjectVariable):

    pass
